from __future__ import annotations

from typing import Any

from aiida import orm
from aiida.engine import ExitCode

from aiida_vasp.parsers.content_parsers import *

from .vasp import MissingFileError, NotificationComposer, QuantityMissingError, VaspParser, get_structure_node

DEFAULT_EXCLUDED_QUANTITIES = (
    'energies',
    'chgcar',
    'wavecar',
    'projectors',
    'charge_density',
    'magnetization_density',
    'elastic_moduli',
    'symmetries',
)

DEFAULT_EXCLUDED_NODE = tuple(['bands', 'dos', 'kpoints', 'trajectory'])

DEFAULT_REQUIRED_QUANTITIES = ('run_status', 'run_stats')

DEFAULT_FILE_MAPPING = {
    'vasprun.xml': 'vasprun.xml',
    'vasp_output': 'vasp_output',
    'OUTCAR': 'OUTCAR',
    'CONTCAR': 'CONTCAR',
    'CHGCAR': 'CHGCAR',
    'IBZKPT': 'IBZKPT',
}
MISC_QUANTITIES = (
    'total_energies',
    'notifications',
    'run_status',
    'run_stats',
    'version',
    'forces',
    'stress',
    'site_magnetization',
    'band_properties',
    'elastic_moduli',
    'symmetries',
    'fermi_level',
    'band_properties',
    'magnetization',
    'neb_data',
    'outcar_positions',
    'outcar_cell',
)


class NebParser(VaspParser):
    """
    Parser for handling NEB calculations.
    """

    def parse(self, **kwargs: Any) -> ExitCode | None:
        user_config = self._init_user_settings()
        # Clear the critical objects list as they do not apply for NEB
        user_config.critical_objects = []

        exit_code = self._get_quantities_to_parse()
        if exit_code is not None:
            return exit_code

        # Check the number of images
        nimages = self.get_num_images()
        self.neb_indices = [f'{i:02d}' for i in range(1, nimages + 1)]

        # Parse the files
        def parse_and_add(
            name: str,
            parser_cls: Any,
            index: Any,
            required: bool = True,
            open_mode: str = 'r',
            content_parser_settings: dict | None = None,
        ) -> None:
            """
            Parse the target file and add the result to the quantities_each dictionary
            For NEB calculations, the results are in the individual folders named with the image index.
            The parsed quantities are rested inside a dictionary with the image index as the key.

            """
            resolved_name = user_config.file_mapping[name]
            # The stdout of the images are in the individual folders with the name stdout
            if name == 'vasp_output' and int(index) > 1:
                resolved_name = 'stdout'
            fpath = index + '/' + resolved_name
            # With the exception that the stdout of the first image is in the base calculation folder
            if name == 'vasp_output' and int(index) == 1:
                fpath = 'stdout'
            try:
                with self.retrieved.open(fpath, open_mode) as handler:
                    try:
                        parser: BaseFileParser = parser_cls(handler=handler, settings=content_parser_settings)
                    except Exception as error:
                        self.errored_parsers[name] = error
                        return
                    if parser.parser_notifications:
                        self.parser_notifications.update(parser.parser_notifications)
                    # Create the dictionary if it doesn't exist
                    if name not in self.quantities_each:
                        self.quantities_each[name] = {}
                    quantities, errored = parser.get_all_quantities()
                    # Save the parsed quantities in the dictionary with index as the key
                    for key, value in quantities.items():
                        if key not in self.quantities_each[name]:
                            self.quantities_each[name][key] = {}
                        self.quantities_each[name][key][index] = value
                    self.errored_quantities.update(errored)
            except FileNotFoundError as error:
                raise MissingFileError(f'{fpath} is missing in the retrieved folder.') from error

        for index in self.neb_indices:
            # We do not parse vasprun.xml as it is not always present in the sub folders
            # TODO - add XDATACAR support
            parse_and_add('OUTCAR', VtstNebOutcarParser, index, required=True)
            parse_and_add('vasp_output', StreamParser, index, required=True)
            parse_and_add('CONTCAR', PoscarParser, index, required=True)

        exit_code = self._post_process_quantities()
        if exit_code is not None:
            return exit_code

        return self._create_outputs()

    def get_num_images(self) -> int:
        """
        Return the number of images
        """
        try:
            nimages = self.node.inputs.parameters['images']
        except KeyError as no_images:
            raise ValueError('No `images` key defined in inputs - this is really an NEB calculation?') from no_images
        return nimages

    def _create_outputs(self) -> ExitCode | None:
        """Create the outputs"""
        # Create the outputs
        self._failed_to_compose = {}
        for name in ['misc', 'structure', 'trajectory', 'arrays']:
            if name in self.nodes_to_exclude:
                continue
            item = None
            try:
                item = getattr(self, '_compose_' + name)(self.quantities_each)
            except (QuantityMissingError, KeyError, ValueError, TypeError) as error:
                self._failed_to_compose[name] = error
                self.logger.warning(f'Failed to compose {name} node: {error}')
                continue
            # If a Node is returned, add it to the outputs
            if isinstance(item, orm.Data):
                self.out(name, item)
            # If a dictionary is returned, it is a namespace output
            if isinstance(item, dict):
                for key, value in item.items():
                    self.out(name + '.' + key, value)
        if (
            any(name in self.user_config.include_node for name in self._failed_to_compose)
            and self.user_config.check_completeness is True
        ):
            return self.exit_codes.ERROR_NOT_ABLE_TO_CREATE_NODE.format(nodes=', '.join(self._failed_to_compose.keys()))
        # Check for errors
        if self.user_config.check_errors is True:
            error = self._check_vasp_errors(self.parser_notifications)
            return error

    def _compose_structure(self, quantities_each: dict[str, Any]) -> dict[str, orm.StructureData]:
        """Compose the `structure` output nodes"""

        data = quantities_each['CONTCAR'].get('structure')
        if len(data) == 0:
            raise QuantityMissingError()

        output = {}
        for index in self.neb_indices:
            output['image_' + index] = get_structure_node(data[index])
        return output

    # At the moment there is no array output for NEB calculations
    # def _compose_arrays(self, quantities_each):
    #     """Generate the generic `arrays` output node"""
    #     array_quantities = ('energies',)
    #     out_arrays = {}
    #     gather_quantities_neb(quantities_each, 'OUTCAR', out_arrays, array_quantities)
    #     # Remove None values in the arrays
    #     out_arrays = {key: value for key, value in out_arrays.items() if value is not None}
    #     if out_arrays:
    #         return orm.ArrayData(out_arrays)
    #     return None

    def _compose_trajectory(self, quantities_each: dict[str, Any]) -> dict[str, orm.TrajectoryData] | None:
        """Compose the `trajectory` output node"""
        output = {}
        for index in self.neb_indices:
            node = orm.TrajectoryData()
            if 'vasprun.xml' in quantities_each:
                traj_data = quantities_each['vasprun.xml'].get('trajectory')
                if traj_data is None:
                    return None
                for key, value in traj_data.items():
                    if key == 'symbols':
                        node.base.attributes.set(key, value)
                    else:
                        node.set_array(key, value)
                output['image_' + index] = node
        return output

    def _compose_misc(self, quantities_each: dict[str, Any]) -> orm.Dict:
        """Compose the `misc` output node"""

        out_dict = {}
        gather_quantities_neb(quantities_each, self.user_config.file_mapping['vasprun.xml'], out_dict, MISC_QUANTITIES)
        gather_quantities_neb(quantities_each, self.user_config.file_mapping['OUTCAR'], out_dict, MISC_QUANTITIES)
        gather_quantities_neb(quantities_each, self.user_config.file_mapping['vasp_output'], out_dict, MISC_QUANTITIES)
        return orm.Dict(dict=out_dict)

    def _check_vasp_errors(self, parser_notifications: dict[str, Any]) -> ExitCode | None:
        """
        Detect simple vasp execution problems and returns the exit_codes to be set
        """
        quantities = {}
        for key, value in self.quantities_each.items():
            for key_, value_ in value.items():
                quantities[key_] = value_
        if 'run_status' not in quantities:
            return self.exit_codes.ERROR_DIAGNOSIS_OUTPUTS_MISSING
        for index in self.neb_indices:
            run_status = quantities['run_status'][index]

            try:
                # We have an overflow in the XML file which is critical, but not reported by VASP in
                # the standard output, so checking this here.
                if parser_notifications.get('vasprun_xml_overflow'):
                    return self.exit_codes.ERROR_OVERFLOW_IN_XML
            except AttributeError:
                pass

            # Return errors related to execution and convergence problems.
            # Note that the order is important here - if a calculation is not finished, we cannot
            # comment on wether properties are converged are not.
            if run_status['finished'] is False:
                return self.exit_codes.ERROR_DID_NOT_FINISH

            if run_status['electronic_converged'] is False:
                return self.exit_codes.ERROR_ELECTRONIC_NOT_CONVERGED

            # Check the ionic convergence issues
            if run_status['ionic_converged'] is False:
                if self.user_config.check_ionic_convergence is True:
                    return self.exit_codes.ERROR_IONIC_NOT_CONVERGED
                self.logger.warning(
                    'The ionic relaxation is not converged, but the calculation is treated as successful.'
                )

        # Check for the existence of critical warnings
        if 'notifications' in quantities:
            notifications = quantities['notifications']['01']
            ignore_all = self.user_config.ignore_notification_errors
            if not ignore_all:
                composer = NotificationComposer(
                    notifications,
                    quantities['run_status']['01'],
                    self.node.inputs,
                    self.exit_codes,
                    critical_notifications=self.user_config.critical_notification_errors,
                )
                exit_code = composer.compose()
                if exit_code is not None:
                    return exit_code
        else:
            self.logger.warning('WARNING: missing notification output for VASP warnings and errors.')

        return None


def gather_quantities_neb(
    quantities_each: dict[str, Any], namespace: str, dst: dict[str, Any], fields: list[str]
) -> None:
    """
    Gather quantities and put them into the target dictionary
    """
    for key, value in quantities_each.get(namespace, {}).items():
        if key in fields:
            dst[key] = value



"""
General utils.

Contains general utils that is not directly coupled to the plugin or AiiDA.
"""

from __future__ import annotations

import os
import shutil
from typing import Callable


def copytree(
    src: str,
    dst: str,
    symlinks: bool = False,
    ignore: Callable[[str, list[str]], list[str]] | None = None,
) -> None:
    """
    Fixes annoying complaint about existing directory running tests

    In Python 3.8 there is a flag for shutil.copytree that handles this,
    but we still support 3.6 and 3.7.
    """
    for item in os.listdir(src):
        source = os.path.join(src, item)
        destination = os.path.join(dst, item)
        if os.path.isdir(source):
            shutil.copytree(source, destination, symlinks, ignore)
        else:
            shutil.copy2(source, destination)



"""
Parser module for composing output of a VASP calculation.

The simplified parser outputs the following nodes:

1. A `misc` node that stores simple summary information such as total energies,
    total run times, any warnings issues during the calculation and if the calculation
    was finished.
2. A `arrays` node that for storing miscellaneous quantities that are arrays by nature and typically
    have a large size.
3. A `trajectory` node for storing the trajectory of geometry optimisation and AIMD.
4. A `bands` node for storing the band structure.
5. A `dos` node for storing the density of states.
6. Other nodes for storing other relevant quantities such as the born effective charges.


Main difference from the previous version

1. `pydantic` is used to validate the parser settings *at submission* time.
2. When parsing retrieved data, we take a 'parse as much as possible' approach -
   the quantities are always parsed if available, and only excluded during the stage of composing the output nodes.
   This is simpler from the previous 'parse only when needed' approach, where multiple checks have be done to work
   out which parser to call and which quantities to include.
3. All parser logic is contained in a single class and can be extended by updating content parsers and modify the
   default settings and update/add the `_compose_xx` methods.
"""

from typing import Any, Dict, List

import numpy as np
from aiida import orm
from aiida.engine import ExitCode
from aiida.parsers.parser import Parser
from pydantic import Field

from aiida_vasp.data.chargedensity import ChargedensityData
from aiida_vasp.data.wavefun import WavefunData
from aiida_vasp.parsers.content_parsers import *
from aiida_vasp.utils.opthold import OptionContainer


class ParserError(RuntimeError):
    pass


class QuantityMissingError(ParserError):
    pass


class RequiredQuantityMissingError(ParserError):
    pass


class MissingFileError(ParserError):
    pass


DEFAULT_EXCLUDED_QUANTITIES = (
    'elastic_moduli',
    'symmetries',
    'parameters',  # The parameters used for the calculation
)

DEFAULT_EXCLUDED_NODE = ('bands', 'dos', 'kpoints', 'trajectory', 'energies', 'wavecar', 'chgcar', 'projectors')

DEFAULT_REQUIRED_QUANTITIES = ('run_status', 'run_stats')

DEFAULT_FILE_MAPPING = {
    'vasprun.xml': 'vasprun.xml',
    'vasp_output': 'vasp_output',
    'OUTCAR': 'OUTCAR',
    'CONTCAR': 'CONTCAR',
    'CHGCAR': 'CHGCAR',
    'IBZKPT': 'IBZKPT',
}
MISC_QUANTITIES = (
    'total_energies',
    'notifications',
    'run_status',
    'run_stats',
    'version',
    'forces',
    'stress',
    'site_magnetization',
    'band_properties',
    'elastic_moduli',
    'symmetries',
    'fermi_level',
    'band_properties',
    'magnetization',
    'parameters',
)

ALLOW_EMPTY = ('notifications',)

# Quantities that should be stored inside separate nodes
STANDALONE_ARRAY_QUANTITIES = {
    'born_charges': 'vasprun.xml',
    'dielectrics': 'vasprun.xml',
    'dynmat': 'vasprun.xml',
    'hessian': 'vasprun.xml',
    'projectors': 'vasprun.xml',
    'energies': 'vasprun.xml',
}


class ParserSettingsConfig(OptionContainer):
    """
    Settings for the VASP parser.
    """

    include_quantity: List[str] = Field(description='Properties to include', default_factory=lambda: [])
    exclude_quantity: List[str] = Field(description='Quantities to be excluded', default_factory=lambda: [])
    required_quantity: List[str] = Field(
        description='Quantities that most be present', default_factory=lambda: list(DEFAULT_REQUIRED_QUANTITIES)
    )
    include_node: List[str] = Field(description='Output node to include', default_factory=lambda: [])
    exclude_node: List[str] = Field(description='Output node to exclude', default_factory=lambda: [])
    file_mapping: Dict[str, str] = Field(
        description='Mapping of file names to quantities', default_factory=lambda: dict(DEFAULT_FILE_MAPPING)
    )
    kpoints_from_ibzkpt: bool = False
    check_completeness: bool = True
    electronic_step_energies: bool = False
    energy_type: List[str] = Field(
        description='Energy types to include', default_factory=lambda: ['energy_extrapolated']
    )
    keep_stream_history: bool = Field(
        description='Whether to keep the history of all notifications in the parsed stream (stdout)', default=False
    )
    ignore_notification_errors: bool = Field(
        description='Whether to ignore errors in the notifications parsed from vasp_output', default=False
    )
    critical_notification_errors: List[str] = Field(
        description='Critical stream errors to raise',
        default_factory=lambda: [
            'brmix',
            'edwave',
            'brmix',
            'cnormn',
            'denmp',
            'dentet',
            'edddav_zhegv',
            'eddrmm_zhegv',
            'edwav',
            'fexcp',
            'fock_acc',
            'non_collinear',
            'not_hermitian',
            'pzstein',
            'real_optlay',
            'rhosyg',
            'rspher',
            'set_indpw_full',
            'sgrcon',
            'no_potimm',
            'magmom',
            'bandocc',
        ],
    )
    critical_objects: List[str] = Field(
        description='Critical objects to be present', default_factory=lambda: ['vasprun.xml', 'OUTCAR']
    )
    check_errors: bool = Field(description='Whether to check for errors in calculation', default=True)
    check_ionic_convergence: bool = Field(
        description='Whether to check for convergence during the relaxation based on the INCAR settings', default=True
    )
    omit_structure: bool = Field(
        description='Whether to omit the structure node from the output if no ionic movement', default=True
    )


class VaspParser(Parser):
    """Class for parsing VASP output files and storing the results in AiiDA."""

    def __init__(self, node: orm.CalcJobNode) -> None:
        """
        Initialize the Parser instance
        """
        super(VaspParser, self).__init__(node)
        # Create the containers
        self.user_config = None
        self.quantities_each: Dict[str, Any] = {}
        self.errored_quantities: Dict[str, Any] = {}
        self.errored_parsers: Dict[str, Any] = {}
        self.parser_notifications: Dict[str, List[str]] = {}
        self.retrieve_object_names: List[str] = []
        self.quantities_to_exclude: List[str] = []
        self.nodes_to_exclude: List[str] = []

    def _init_user_settings(self) -> ParserSettingsConfig:
        """Initialize the settings from the inputs."""
        if 'settings' in self.node.inputs:
            user_config: ParserSettingsConfig = ParserSettingsConfig(
                **self.node.inputs.settings.get_dict().get('parser_settings', {})
            )
        else:
            user_config = ParserSettingsConfig()
        # Initialize the containers
        self.user_config = user_config
        return user_config

    def _get_quantities_to_parse(self) -> ExitCode | None:
        """Return the list of quantities to parse."""
        # Apply the modifiers
        user_config = self.user_config
        quantities_to_exclude = [key for key in DEFAULT_EXCLUDED_QUANTITIES if key not in user_config.include_quantity]
        quantities_to_exclude += user_config.exclude_quantity
        nodes_to_exclude = [key for key in DEFAULT_EXCLUDED_NODE if key not in user_config.include_node]
        nodes_to_exclude += user_config.exclude_node
        self.quantities_to_exclude = quantities_to_exclude
        self.nodes_to_exclude = nodes_to_exclude

        # Check for critical missing objects
        self.retrieve_object_names = self.retrieved.list_object_names()
        missing = False
        for name, _ in user_config.file_mapping.items():
            if name in user_config.critical_objects and name not in self.retrieve_object_names:
                missing = True
        if missing is True:
            return self.exit_codes.ERROR_CRITICAL_MISSING_OBJECT

    def _post_process_quantities(self) -> ExitCode | None:
        """Post-process the parsed quantities."""

        # Warn about errored/missing quantities and parsers

        if self.errored_quantities:
            self.logger.warning(
                f'The following quantities cannot be parsed due to errors: {", ".join(self.errored_quantities)}'
            )
        if self.errored_parsers:
            self.logger.warning(
                f'The following parsers cannot be instantiated due to: {", ".join(self.errored_parsers)}'
            )

        # Remove the quantities
        for name, parsed_quantities in self.quantities_each.items():
            for sub_key in list(parsed_quantities.keys()):
                if sub_key in self.quantities_to_exclude:
                    del parsed_quantities[sub_key]

        # Check in required quantities are present
        missing_required = []
        for name in self.user_config.required_quantity:
            exists = False
            for _, value in self.quantities_each.items():
                if value.get(name) is not None:
                    exists = True
                    break
            if exists is False:
                missing_required.append(name)
        if missing_required:
            return self.exit_codes.ERROR_NOT_ABLE_TO_PARSE_QUANTITY.format(quantity=','.join(missing_required))

    def parse(self, **kwargs: Any) -> ExitCode | None:
        """
        Parse outputs, store results in database.
        """
        user_config = self._init_user_settings()

        exit_code = self._get_quantities_to_parse()
        if exit_code is not None:
            return exit_code

        # Parse the files
        def parse_and_add(
            name: str,
            parser_cls: Any,
            required: bool = True,
            open_mode: str = 'r',
            content_parser_settings: dict | None = None,
        ) -> None:
            """Parse the target file and add the result to the quantities_each dictionary"""
            resolved_name = user_config.file_mapping[name]
            if resolved_name in self.retrieve_object_names:
                with self.retrieved.open(resolved_name, open_mode) as handler:
                    try:
                        parser: BaseFileParser = parser_cls(handler=handler, settings=content_parser_settings)
                    except Exception as error:
                        self.errored_parsers[name] = error
                        return
                    if parser.parser_notifications:
                        self.parser_notifications.update(parser.parser_notifications)
                    self.quantities_each[name], errored = parser.get_all_quantities()
                    self.errored_quantities.update(errored)
            elif user_config.check_completeness is True and required is True:
                raise MissingFileError(f'{resolved_name} is missing in the retrieved folder.')

        parse_and_add(
            'vasprun.xml',
            VasprunParser,
            required=True,
            open_mode='rb',
            content_parser_settings={
                'electronic_step_energies': user_config.electronic_step_energies,
                'energy_type': user_config.energy_type,
                'stream_history': user_config.keep_stream_history,
            },
        )
        parse_and_add('OUTCAR', OutcarParser, required=True)
        parse_and_add('vasp_output', StreamParser, required=True)
        parse_and_add('CONTCAR', PoscarParser, required=True)

        if user_config.kpoints_from_ibzkpt:
            parse_and_add('IBZKPT', KpointsParser, required=True)

        exit_code = self._post_process_quantities()
        if exit_code is not None:
            return exit_code

        return self._create_outputs()

    def _create_outputs(self) -> ExitCode | None:
        """Create the output nodes"""
        # Create the outputs
        self._failed_to_compose = {}

        # Call the _compose_xx methods to create the output nodes
        for method_name in [item for item in self.__dir__() if item.startswith('_compose_')]:
            name = method_name.replace('_compose_', '')
            if name in self.nodes_to_exclude:
                continue
            node_or_dict = None
            try:
                node_or_dict = getattr(self, '_compose_' + name)(self.quantities_each)
            except (QuantityMissingError, KeyError, ValueError, TypeError, AttributeError) as error:
                self._failed_to_compose[name] = error
                self.logger.warning(f'Failed to compose {name} node: {error}')
                continue
            if isinstance(node_or_dict, orm.Data):
                self.out(name, node_or_dict)
            elif isinstance(node_or_dict, dict):
                for key, value in node_or_dict.items():
                    self.out(key, value)
        if (
            any(name in self.user_config.include_node for name in self._failed_to_compose)
            and self.user_config.check_completeness is True
        ):
            return self.exit_codes.ERROR_NOT_ABLE_TO_CREATE_NODE.format(nodes=', '.join(self._failed_to_compose.keys()))
        # Check for errors
        if self.user_config.check_errors is True:
            error = self._check_vasp_errors(self.parser_notifications)
            return error

    def _compose_misc(self, quantities_each: dict[str, Any]) -> orm.Dict:
        """Compose the `misc` output node"""

        out_dict = {}
        gather_quantities(quantities_each, self.user_config.file_mapping['vasprun.xml'], out_dict, MISC_QUANTITIES)
        gather_quantities(quantities_each, self.user_config.file_mapping['OUTCAR'], out_dict, MISC_QUANTITIES)
        gather_quantities(quantities_each, self.user_config.file_mapping['vasp_output'], out_dict, MISC_QUANTITIES)
        # Filter field with all empty container
        out_dict = {key: value for key, value in out_dict.items() if not is_all_empty(value) or key in ALLOW_EMPTY}
        return orm.Dict(dict=out_dict)

    def _compose_structure(self, quantities_each: dict[str, Any]) -> orm.StructureData | None:
        """Compose the `structure` output node"""

        data = None
        # Omit output structure if not doing ionic relaxation
        # Better to inspect the parameters recorded directly inside the vasprun.xml
        if 'parameters' in self.node.inputs:
            incar_dict = {key.lower(): value for key, value in self.node.inputs.parameters.get_dict().items()}
            if (
                incar_dict.get('ibrion', -1) < 0 or incar_dict.get('nsw', 0) <= 0
            ) and self.user_config.omit_structure is True:
                self.logger.info('No ionic movement detected, omitting the structure output node.')
                return None

        if 'vasprun.xml' in quantities_each:
            data = quantities_each['vasprun.xml'].get('structure')
        if data is None:
            data = quantities_each.get('CONTCAR', {}).get('structure')
        if data is None:
            raise QuantityMissingError()
        return get_structure_node(data)

    def _compose_wavecar(self, quantities_each: dict[str, Any]) -> None:
        """Compose the `wavecar` output node"""

        # Check if WAVECAR is present in the retrieved folder
        if 'WAVECAR' in self.retrieve_object_names:
            with self.retrieved.base.repository.open('WAVECAR', 'rb') as handler:
                self.outputs['wavecar'] = WavefunData(file=handler, filename='WAVECAR')
        else:
            self.logger.warning('WAVECAR is not present in the retrieved folder.')

    def _compose_chgcar(self, quantities_each: dict[str, Any]) -> None:
        """Compose the `chgcar` output node"""

        # Check if WAVECAR is present in the retrieved folder
        if 'CHGCAR' in self.retrieve_object_names:
            with self.retrieved.base.repository.open('CHGCAR', 'rb') as handler:
                self.outputs['chgcar'] = ChargedensityData(file=handler, filename='CHGCAR')
        else:
            self.logger.warning('CHGCAR is not present in the retrieved folder.')

    def _compose_arrays(self, quantities_each: dict[str, Any]) -> dict[str, orm.ArrayData]:
        """Generate the generic `arrays` output node"""
        out_arrays = {}

        # Compose the standalone arrays - each corresponds to a single quantity
        for name, file_name in STANDALONE_ARRAY_QUANTITIES.items():
            if name in self.nodes_to_exclude:
                continue
            array_node = self._make_standalone_array(quantities_each, name, file_name)
            if array_node is not None:
                out_arrays[name] = array_node
        return out_arrays

    def _make_standalone_array(
        self, quantities_each: dict[str, Any], name: str, file_name: str = 'vasprun.xml'
    ) -> orm.ArrayData | None:
        """Compose the `dielectrics` output node"""
        # The output can be an array or a dictionary of arrays - both cases should be handled
        arrays_or_dict = quantities_each.get(file_name, {}).get(name)
        # Avoid creating empty arrays
        if isinstance(arrays_or_dict, dict) and len(arrays_or_dict) > 0:
            arrays_or_dict = {key: value for key, value in arrays_or_dict.items() if value is not None}
            return orm.ArrayData(arrays_or_dict)
        elif isinstance(arrays_or_dict, (np.ndarray, list)) and len(arrays_or_dict) > 0:
            return orm.ArrayData({name: arrays_or_dict})
        return None

    def _compose_kpoints(self, quantities_each: dict[str, Any]) -> orm.KpointsData:
        """Compose the `kpoints` output node"""
        kpoints_data = None
        if self.user_config.kpoints_from_ibzkpt is True:
            kpoints_data = quantities_each['IBZKPT']['kpoints']
        elif 'vasprun.xml' in quantities_each:
            kpoints_data = quantities_each['vasprun.xml'].get('kpoints')

        if kpoints_data is not None:
            node = orm.KpointsData()
            if kpoints_data['mode'] == 'explicit':
                node.set_kpoints(
                    kpoints_data['points'], weights=kpoints_data['weights'], cartesian=kpoints_data['cartesian']
                )
            elif kpoints_data['mode'] == 'automatic':
                node.set_kpoints_mesh(kpoints_data['divisions'], offset=kpoints_data['shifts'])
            else:
                raise ValueError(f'Unknown kpoints mode {kpoints_data["mode"]}')
            # Record the cell for which the kpoints are defined for
            node.set_cell(quantities_each['vasprun.xml']['structure']['unitcell'])
            return node
        raise QuantityMissingError('No valid kpoints data to use')

    def _compose_trajectory(self, quantities_each: dict[str, Any]) -> orm.TrajectoryData | None:
        """Compose the `trajectory` output"""

        if 'vasprun.xml' in quantities_each:
            node = orm.TrajectoryData()
            traj_data = quantities_each['vasprun.xml'].get('trajectory')
            # No need to carry on if there are no trajectory data
            if traj_data is None or len(traj_data) == 0:
                return None
            for key, value in traj_data.items():
                if key == 'symbols':
                    node.base.attributes.set(key, value)
                elif value.dtype.hasobject:
                    self.logger.warning(f'Cannot set array {key}: {value} in TrajectoryData as it is not numerical.')
                else:
                    node.set_array(key, value)
            for key, value in quantities_each['vasprun.xml']['energies'].items():
                node.set_array(key, value)
            return node
        return None

    def _compose_bands(self, quantities_each: dict[str, Any]) -> orm.BandsData:
        """Compose the `band` node"""
        if 'vasprun.xml' in quantities_each:
            deigen = quantities_each['vasprun.xml']['eigenvalues']
            docc = quantities_each['vasprun.xml']['occupancies']
            if 'total' in deigen:
                eigenvalues = np.array(deigen['total'])
                occupancies = np.array(docc['total'])
            else:
                eigenvalues = np.array([deigen['up'], deigen['down']])
                occupancies = np.array([docc['up'], docc['down']])
            node = orm.BandsData()
            kpoints = self._compose_kpoints(quantities_each)
            node.set_kpointsdata(kpoints)
            node.set_bands(eigenvalues, occupations=occupancies)

            # Record the Fermi level if available
            node.base.attributes.set('fermi_level', quantities_each['vasprun.xml'].get('fermi_level'))
            node.base.attributes.set('efermi', quantities_each['vasprun.xml'].get('fermi_level'))
            node.set_cell(quantities_each['vasprun.xml']['structure']['unitcell'])
            return node

    def _compose_dos(self, quantities_each: dict[str, Any]) -> orm.ArrayData | None:
        """Compose the `dos` node"""
        arrays_dict = {}
        if 'vasprun.xml' in quantities_each:
            gather_quantities(quantities_each, 'dos', arrays_dict, ['dos'], flatten_dict=True)
        if arrays_dict:
            node = orm.ArrayData(arrays_dict['dos'])
            return node

    def _check_vasp_errors(self, parser_notifications: dict[str, Any]) -> ExitCode | None:
        """
        Detect simple vasp execution problems and returns the exit_codes to be set
        """
        quantities = {}
        for key, value in self.quantities_each.items():
            for key_, value_ in value.items():
                quantities[key_] = value_
        if 'run_status' not in quantities:
            return self.exit_codes.ERROR_DIAGNOSIS_OUTPUTS_MISSING
        run_status = quantities['run_status']

        try:
            # We have an overflow in the XML file which is critical, but not reported by VASP in
            # the standard output, so checking this here.
            if parser_notifications.get('vasprun_xml_overflow'):
                return self.exit_codes.ERROR_OVERFLOW_IN_XML
        except AttributeError:
            pass

        # Return errors related to execution and convergence problems.
        # Note that the order is important here - if a calculation is not finished, we cannot
        # comment on wether properties are converged are not.
        if run_status['finished'] is False:
            return self.exit_codes.ERROR_DID_NOT_FINISH

        if run_status['electronic_converged'] is False:
            return self.exit_codes.ERROR_ELECTRONIC_NOT_CONVERGED

        # Check the ionic convergence issues
        if run_status['ionic_converged'] is False:
            if self.user_config.check_ionic_convergence is True:
                return self.exit_codes.ERROR_IONIC_NOT_CONVERGED
            self.logger.warning('The ionic relaxation is not converged, but the calculation is treated as successful.')

        # Check for the existence of critical warnings
        if 'notifications' in quantities:
            notifications = quantities['notifications']
            ignore_all = self.user_config.ignore_notification_errors
            if not ignore_all:
                composer = NotificationComposer(
                    notifications,
                    quantities['run_status'],
                    self.node.inputs,
                    self.exit_codes,
                    critical_notifications=self.user_config.critical_notification_errors,
                )
                exit_code = composer.compose()
                if exit_code is not None:
                    return exit_code
        else:
            self.logger.warning('WARNING: missing notification output for VASP warnings and errors.')

        return None


def gather_quantities(
    quantities_each: dict[str, Any], namespace: str, dst: dict[str, Any], fields: list[str], flatten_dict: bool = False
) -> None:
    """
    Gather quantities and put them into the target dictionary
    """
    for key, value in quantities_each.get(namespace, {}).items():
        if key in fields:
            if isinstance(value, dict) and flatten_dict:
                # flatten the dictionary - prepend the key with the name of the quantity
                for key2, value2 in value.items():
                    dst[key + '_' + key2] = value2
            else:
                dst[key] = value


class NotificationComposer:
    """Compose errors codes based on the notifications"""

    def __init__(
        self,
        notifications: list[dict[str, Any]],
        run_status: dict[str, Any],
        inputs: dict[str, Any],
        exit_codes: Any,
        critical_notifications: list[str],
    ) -> None:
        """
        Composed error codes based on the notifications

        Some of the errors need to have additional properties inspected before they can be emitted,
        as they might be trigged in a harmless way.

        To add new checkers, one needs to implement a property with the name of the error for this class and
        contains the code for checking. This property should return the exit_code or return None. The property
        is inspected if its name is in the list critical notifications.
        """
        self.notifications = notifications
        self.notifications_dict = {item['name']: item['message'] for item in self.notifications}
        self.run_status = run_status
        self.inputs = inputs
        self.exit_codes = exit_codes
        self.critical_notifications = critical_notifications

    def compose(self) -> ExitCode | None:
        """
        Compose the exit codes

        Returns None if no exit code should be emitted, otherwise emit the error code.
        """

        for critical in self.critical_notifications:
            # Check for any special handling
            if hasattr(self, critical):
                output = getattr(self, critical)
                if output:
                    return output
            # No special handling, just check if it exists
            elif critical in self.notifications_dict:
                return self.exit_codes.ERROR_VASP_CRITICAL_ERROR.format(error_message=self.notifications_dict[critical])
        return None

    @property
    def brmix(self) -> ExitCode | None:
        """Check if BRMIX should be emitted"""
        if 'brmix' not in self.notifications_dict:
            return None

        # If NELECT is set explicitly for the calculation then this is not an critical error
        if 'parameters' in self.inputs and 'nelect' in self.inputs['parameters'].get_dict():
            return None

        return self.exit_codes.ERROR_VASP_CRITICAL_ERROR.format(error_message=self.notifications_dict['brmix'])

    @property
    def edddav_zhegv(self) -> ExitCode | None:
        """Check if EDDDAV call to ZHEGV should be emitted. Sometimes it has converged."""
        if 'edddav_zhegv' not in self.notifications_dict:
            return None

        if self.run_status['electronic_converged']:
            return None

        return self.exit_codes.ERROR_VASP_CRITICAL_ERROR.format(error_message=self.notifications_dict['edddav_zhegv'])

    @property
    def eddrmm_zhegv(self) -> ExitCode | None:
        """Check if EDDRMM call to ZHEGV should be emitted. Sometimes it has converged."""
        if 'eddrmm_zhegv' not in self.notifications_dict:
            return None

        if self.run_status['electronic_converged']:
            return None

        return self.exit_codes.ERROR_VASP_CRITICAL_ERROR.format(error_message=self.notifications_dict['eddrmm_zhegv'])


def get_structure_node(structure_dict: dict[str, Any]) -> orm.StructureData:
    """Compose a structure node from the dictionary output by the parser"""
    node = orm.StructureData()
    node.set_cell(structure_dict['unitcell'])
    for site in structure_dict['sites']:
        node.append_atom(position=site['position'], symbols=site['symbol'], name=site['kind_name'])
    return node


def is_all_empty(obj: dict | list) -> bool:
    """Check if all elements of a dictionary or list are empty"""
    if isinstance(obj, dict):
        if len(obj) == 0:
            return True
        else:
            return all(is_all_empty(value) for value in obj.values())
    elif isinstance(obj, list):
        if len(obj) == 0:
            return True
        else:
            return all(is_all_empty(value) for value in obj)
    else:
        return False



"""
Fixtures for workchains.

------------------------
Fixtures representing WorkChains, for usage in tests.
This can be either Top level workchains using mock-ups of low level WorkChains
for testing complex WorkChains, or the mock-ups themselves.
"""
# pylint: disable=too-few-public-methods, redefined-outer-name, import-outside-toplevel

import pytest
from aiida.plugins import WorkflowFactory


@pytest.fixture
def mock_base_workchain():
    """Fixture for a mock-up of the VaspWorkChain."""
    _base_wc_cls = WorkflowFactory('vasp.vasp')

    class VaspWorkChain(_base_wc_cls):
        pass

    return VaspWorkChain


@pytest.fixture
def mock_converge_workchain(mock_base_workchain):  # pylint: disable=unused-argument
    """Fixture for a ConvergenceWorkChain using a mock-up for the lower level VaspWorkChain."""
    _base_wc_cls = WorkflowFactory('vasp.converge')

    class ConvergeWorkCHain(_base_wc_cls):
        _next_workchain = mock_base_workchain

    return ConvergeWorkCHain


def mock_factory(base_class, run_method):
    """
    Return a mock-up of a class inheriting from CalcJobNode.

    :param base_class: THe base class inheriting from CalcJobNode as e.g. defined in a plugin.
    :param run_method: a method with signature run(self) that will add outputs based on the inputs to the CalcJobNode.
    """

    class MockCalcJob(base_class):
        """A mock CalcJob class."""

        run = run_method

        @classmethod
        def define(cls, spec):
            super(base_class, cls).define(spec)  # pylint: disable=bad-super-call
            spec.outline(cls.run)

    return MockCalcJob



"""Test the INCAR parser."""
# pylint: disable=redefined-outer-name, unused-wildcard-import, unused-argument, wildcard-import

import pytest
from aiida import orm

from aiida_vasp.parsers.content_parsers.incar import IncarParser

compare_incar = {'gga': 'PE', 'gga_compat': False, 'lorbit': 11, 'magmom': '30 * 2*0.', 'sigma': 0.5}


@pytest.mark.parametrize(['incar_parser'], [('incar',)], indirect=True)
def test_parse_incar(incar_parser):
    """Load a reference INCAR parser.

    We check that it parses and provides the correct content for the default INCAR.

    """

    # The structure for the INCAR parser should have the key `incar-structure`
    result = incar_parser.get_quantity('incar')
    assert result == compare_incar


@pytest.mark.parametrize(['incar_parser'], [('phonondb',)], indirect=True)
def test_parse_incar_phonon(incar_parser):
    """Load a reference INCAR parser.

    We check that it parses and provides the correct content for an INCAR used for
    phonon calculations.

    """

    incar = incar_parser.incar
    assert incar['prec'] == 'Accurate'
    assert incar['ibrion'] == -1
    assert incar['encut'] == pytest.approx(359.7399)
    assert incar['lreal'] is False


@pytest.mark.parametrize(['incar_parser'], [('incar',)], indirect=True)
def test_parse_incar_write(incar_parser, tmpdir):
    """Load a reference INCAR parser and check that the write functionality works.

    Here we make sure the write function of the content parser works.

    """
    # Write the loaded structure to file
    temp_path = str(tmpdir.join('INCAR'))
    incar_parser.write(temp_path)

    # Load the written structure using a new content parser instance
    content = None
    with open(temp_path, 'r', encoding='utf8') as handler:
        content = handler.readlines()
    ref_content = ['GGA = PE\n', 'GGA_COMPAT = .FALSE.\n', 'LORBIT = 11\n', 'MAGMOM = 30 * 2*0.\n', 'SIGMA = 0.5\n']
    assert content == ref_content


def test_parse_incar_data(vasp_params, tmpdir):
    """Load a reference AiiDA Dict and check that the parser can
    initialize using the data.

    Using the Dict sitting in the initialized parser it should
    write that content to an INCAR file when issuing write which is also tested,
    file is reloaded and content checked.

    """

    # Initialize parser with an existing reference Dict
    incar_parser = IncarParser(data=vasp_params)

    # Check that get_quantity return the same Dict instance
    assert vasp_params == incar_parser.get_quantity('key_does_not_matter')

    # Write the loaded Dict to file, which behind the scenes convert it
    # to a INCAR format
    temp_path = str(tmpdir.join('INCAR'))
    incar_parser.write(temp_path)

    # Load the written INCAR using a new content parser instance
    parser = None
    with open(temp_path, 'r', encoding='utf8') as handler:
        parser = IncarParser(handler=handler)
    result = parser.get_quantity('incar')
    assert vasp_params.get_dict() == result


@pytest.mark.parametrize(['incar_parser'], [(['incar', 'INCAR.nohash'],)], indirect=True)
def test_parse_incar_nohash(incar_parser):
    """Load a reference INCAR parser.

    Using parsevasp. Returned content should be None
    since parsevasp refuse to parse an INCAR where the
    comments does not start with hashtags.

    """

    result = incar_parser.incar
    assert result is None


def test_parse_incar_invalid_tag(vasp_params, tmpdir):
    """Test parsing an INCAR with an invalid tag."""
    params = vasp_params.get_dict()
    params.update(foo='bar')
    vasp_params_modified = orm.Dict(dict=params)
    parser = IncarParser(data=vasp_params_modified)
    temp_path = str(tmpdir.join('INCAR'))
    with pytest.raises(SystemExit):
        parser.write(temp_path)



from .neb import VaspNEBWorkChain
from .v2.bands import VaspBandsWorkChain, VaspHybridBandsWorkChain
from .v2.common.builder_updater import *
from .v2.converge import VaspConvergenceWorkChain
from .v2.relax import VaspMultiStageRelaxWorkChain, VaspRelaxWorkChain
from .v2.vasp import VaspWorkChain

__all__ = (
    'VaspBandUpdater',
    'VaspBandsWorkChain',
    'VaspBuilderUpdater',
    'VaspConvUpdater',
    'VaspConvergenceWorkChain',
    'VaspHybridBandUpdater',
    'VaspHybridBandsWorkChain',
    'VaspMultiStageRelaxWorkChain',
    'VaspNEBUpdater',
    'VaspNEBWorkChain',
    'VaspPresetConfig',
    'VaspRelaxUpdater',
    'VaspRelaxWorkChain',
    'VaspWorkChain',
)



"""
Module to provide dryrun functionality.
"""

from __future__ import annotations

import shutil
from math import ceil, gcd
from typing import Optional
from warnings import warn

import numpy as np
from aiida.common.folders import SubmitTestFolder
from aiida.engine.daemon.execmanager import upload_calculation
from aiida.engine.processes.builder import ProcessBuilder
from aiida.orm import Dict, KpointsData
from aiida.transports.plugins.local import LocalTransport

from aiida_vasp.assistant.parameters import ParametersMassage
from aiida_vasp.calcs.vasp import VaspCalculation
from aiida_vasp.commands.dryrun_vasp import dryrun_vasp as _vasp_dryrun
from aiida_vasp.data.potcar import PotcarData


class JobScheme:
    """
    A class representing the scheme of the jobs.
    """

    def __init__(
        self,
        n_kpoints: int,
        n_procs: int,
        n_nodes: Optional[int] = None,
        cpus_per_node: Optional[int] = None,
        npw: Optional[int] = None,
        nbands: Optional[int] = None,
        ncore_within_node: bool = True,
        ncore_strategy: str = 'maximise',
        wf_size_limit: float = 1000,
    ) -> None:
        """
        Instantiate a JobScheme object.

        :param n_kpoints: Number of kpoints.
        :param n_procs: Number of processes.
        :param n_nodes: Number of nodes.
        :param cpus_per_node: Number of CPUs per node.
        :param npw: Number of plane waves.
        :param nbands: Number of bands.
        :param ncore_within_node: If True, limit plane-wave parallelisation to within each node.
        :param ncore_strategy: Strategy for optimising NCORE, choose from 'maximise' and 'balance'.
        :param wf_size_limit: Limit of the ideal wavefunction size per process in MB.
        """
        self.n_kpoints = n_kpoints
        self.n_procs = n_procs
        self.n_nodes = n_nodes
        self.cpus_per_node = cpus_per_node
        self.npw = npw
        self.nbands = nbands
        self.ncore_within_node = ncore_within_node
        self.ncore_strategy = ncore_strategy
        self.wf_size_limit = wf_size_limit

        self.n_kgroup = None  # KPOINT groups
        self.n_bgroup = None  # Band groups
        self.n_pgroup = None  # Plane wave groups

        self.kpar = None  # Value for the KPAR
        self.npar = None
        self.ncore = None  # Value for the ncore
        self.new_nbands = nbands  # Value for the new nbands
        self.nbands_amplification = None  # Amplification factor for the NBAND round up
        self.ncore_balance = None  # NCORE/NPAR balance factor

        self.solve_kpar()
        self.solve_ncore()

    @classmethod
    def from_dryrun(cls, dryrun_outcome: dict, n_procs: int, **kwargs) -> JobScheme:
        """
        Construct from dryrun results.

        :param dryrun_outcome: The outcome from a dryrun.
        :param n_procs: Number of processes.
        :param kwargs: Additional keyword arguments.

        :returns: A `JobScheme` object
        """
        kwargs['n_kpoints'] = dryrun_outcome.get('num_kpoints')
        kwargs['nbands'] = dryrun_outcome.get('num_bands')
        kwargs['npw'] = dryrun_outcome.get('num_plane_waves')
        kwargs['n_procs'] = n_procs
        return cls(**kwargs)

    def solve_kpar(self) -> int:
        """
        Solve for the optimum strategy for KPAR.

        :returns: The optimized KPAR value.
        """
        kpar = gcd(self.n_kpoints, self.n_procs)
        self.kpar = kpar
        # If we did not set nbands or npw, we cannot adjust KAR to avoid memory issues
        if any(map(lambda x: x is None, [self.nbands, self.npw])):
            warn(
                'Cannot limit KAR for memory requirement without supplying both NBANDS and NPW',
                UserWarning,
            )
            return kpar

        # Reduce the KPAR
        if self.size_wavefunction_per_proc > self.wf_size_limit:
            for candidate in factors(kpar):
                self.kpar = candidate
                if self.size_wavefunction_per_proc < self.wf_size_limit:
                    kpar = candidate
                    break
        if self.size_wavefunction_per_proc > self.wf_size_limit:
            warn(
                ('Expected wavefunction size per process {} MB is larger than the limit {} MB').format(
                    self.size_wavefunction_per_proc, self.wf_size_limit
                ),
                UserWarning,
            )
        return kpar

    @property
    def nk_per_group(self) -> int:
        """Number of kpoints per group."""
        return self.n_kpoints // self.kpar

    @property
    def procs_per_kgroup(self) -> int:
        """Number of processes per kpoint group."""
        return self.n_procs // self.kpar

    def solve_ncore(self) -> int:
        """
        Solve for NCORE.

        :returns: The optimized NCORE value.
        """
        # Cannot solve if no nbands provided or does not know how many cpus per node
        if self.nbands is None:
            return
        if self.ncore_within_node and (self.cpus_per_node is None):
            return

        combs = []
        for ncore in factors(self.procs_per_kgroup):
            if ncore > 12:
                continue
            # Only consider ncore that is a multiple of the cpus per node
            if self.ncore_within_node and self.cpus_per_node % ncore != 0:
                continue
            npar = self.procs_per_kgroup // ncore
            new_nbands = ceil(self.nbands / npar) * npar
            factor = new_nbands / self.nbands
            combs.append((ncore, factor, abs(ncore / npar - 1), new_nbands))

        combs = list(filter(lambda x: x[1] < 1.2, combs))
        if self.ncore_strategy == 'balance':
            combs.sort(key=lambda x: x[2])
        elif self.ncore_strategy == 'maximise':
            combs.sort(key=lambda x: x[0], reverse=True)
        else:
            raise RuntimeError(f'NCORE strategy: <{self.ncore_strategy}> is invalid')

        # Take the maximum ncore
        ncore, factor, balance, new_nbands = combs[0]

        self.ncore = ncore
        self.npar = self.procs_per_kgroup // ncore
        self.nbands_amplification = factor
        self.new_nbands = new_nbands
        self.ncore_balance = balance
        return ncore

    @property
    def size_wavefunction(self) -> float:
        """Memory requirement for the wavefunction in MB."""
        return self.n_kpoints * self.new_nbands * self.npw * 16 / 1048576

    @property
    def size_wavefunction_per_proc(self) -> float:
        """Memory requirement for the wavefunction per process."""
        return self.size_wavefunction / self.procs_per_kgroup


def factors(num: int) -> list[int]:
    """
    Return all factors of a number in descending order, including the number itself.

    :param num: The number to factor.

    :returns: A list of factors.
    """
    result = [num]
    for i in range(num // 2 + 1, 0, -1):
        if num % i == 0:
            result.append(i)
    return result


def dryrun_vasp(
    input_dict: dict | ProcessBuilder,
    vasp_exe: str = 'vasp_std',
    timeout: int = 10,
    work_dir: str | None = None,
    keep: bool = False,
) -> dict:
    """
    Perform a dryrun for a VASP calculation, return obtained information.

    :param input_dict: The input dictionary/builder for `VaspCalculation`.
    :param vasp_exe: The VASP executable to be used.
    :param timeout: Timeout for the underlying VASP process in seconds.
    :param work_dir: Working directory, if not supplied, will use a temporary directory.
    :param keep: Whether to keep the dryrun output.

    :returns: A dictionary of the dry run results parsed from OUTCAR.
    """
    # Deal with passing an process builder
    if isinstance(input_dict, ProcessBuilder):
        try:
            output_node = prepare_inputs(input_dict)
        except Exception as error:
            raise error
    else:
        try:
            output_node = prepare_inputs(input_dict)
        except Exception as error:
            raise error

    folder = output_node.dry_run_info['folder']
    outcome = _vasp_dryrun(folder, vasp_exe=vasp_exe, timeout=timeout, work_dir=work_dir, keep=keep)
    if not keep:
        shutil.rmtree(folder)

    return outcome


def get_jobscheme(input_dict: dict, nprocs: int, vasp_exe: str = 'vasp_std', **kwargs) -> JobScheme:
    """
    Perform a dryrun for the input and work out the best parallelisation strategy.

    :param input_dict: Inputs of the VaspCalculation.
    :param nprocs: Target number of processes to be used.
    :param vasp_exe: The executable of local VASP program to be used.
    :param kwargs: Additional keyword arguments to be passed to `JobScheme`.

    :returns: A `JobScheme` object.
    """
    dryout = dryrun_vasp(input_dict, vasp_exe)
    scheme = JobScheme.from_dryrun(dryout, nprocs, **kwargs)
    return scheme


def prepare_inputs(inputs: dict) -> VaspCalculation:
    """
    Prepare inputs for VASP calculation.

    :param inputs: The inputs to prepare.

    :returns: The prepared inputs.
    """
    inputs = dict(inputs)
    inputs['metadata'] = dict(inputs['metadata'])
    inputs['metadata']['store_provenance'] = False
    inputs['metadata']['dry_run'] = True
    vasp = VaspCalculation(inputs=inputs)

    with LocalTransport() as transport:
        with SubmitTestFolder() as folder:
            calc_info = vasp.presubmit(folder)
            transport.chdir(folder.abspath)
            upload_calculation(
                vasp.node,
                transport,
                calc_info,
                folder,
                inputs=vasp.inputs,
                dry_run=True,
            )
            vasp.node.dry_run_info = {
                'folder': folder.abspath,
                'script_filename': vasp.node.get_option('submit_script_filename'),
            }
    return vasp.node


def dryrun_relax_builder(builder: ProcessBuilder, **kwargs) -> dict:
    """
    Dry run a relaxation workchain builder.

    :param builder: The builder to dry run.
    :param kwargs: Additional keyword arguments.

    :returns: The results of the dry run.
    """
    vasp_builder = VaspCalculation.get_builder()

    # Setup the builder for the bare calculation

    # Convert into CalcJob inputs
    vasp_builder.code = builder.vasp.code
    pdict = builder.vasp.parameters.get_dict()
    parameters_massager = ParametersMassage(pdict, None)

    vasp_builder.parameters = Dict(dict=parameters_massager.parameters.incar)
    if 'dynamics' in parameters_massager.parameters:
        vasp_builder.dynamics = Dict(dict=parameters_massager.parameters.dynamics)

    if builder.vasp.kpoints is not None:
        vasp_builder.kpoints = builder.vasp.kpoints
    else:
        vasp_builder.kpoints = KpointsData()
        vasp_builder.kpoints.set_cell_from_structure(builder.structure)
        vasp_builder.kpoints.set_kpoints_mesh_from_density(builder.vasp.kpoints_spacing.value * np.pi * 2)
    vasp_builder.metadata.options = builder.vasp.options.get_dict()  # pylint: disable=no-member
    vasp_builder.potential = PotcarData.get_potcars_from_structure(
        builder.structure,
        builder.vasp.potential_family.value,
        builder.vasp.potential_mapping.get_dict(),
    )
    vasp_builder.structure = builder.structure

    return dryrun_vasp(vasp_builder, **kwargs)


def dryrun_vaspu_builder(builder: ProcessBuilder, **kwargs) -> dict:
    """
    Dry run a vaspu.vasp workchain builder.

    :param builder: The builder to dry run.
    :param kwargs: Additional keyword arguments.

    :returns: The results of the dry run.
    """
    pdict = builder.parameters.get_dict()
    vasp_builder = VaspCalculation.get_builder()

    parameters_massager = ParametersMassage(pdict, None)

    vasp_builder.parameters = Dict(dict=parameters_massager.parameters.incar)
    if 'dynamics' in parameters_massager.parameters:
        vasp_builder.dynamics = Dict(dict=parameters_massager.parameters.dynamics)

    # Setup the builder for the bare calculation
    vasp_builder.code = builder.code
    if builder.kpoints is not None:
        vasp_builder.kpoints = builder.kpoints
    else:
        vasp_builder.kpoints = KpointsData()
        vasp_builder.kpoints.set_cell_from_structure(builder.structure)
        vasp_builder.kpoints.set_kpoints_mesh_from_density(builder.kpoints_spacing.value * np.pi * 2)
    vasp_builder.metadata.options = builder.options.get_dict()  # pylint: disable=no-member
    vasp_builder.potential = PotcarData.get_potcars_from_structure(
        builder.structure,
        builder.potential_family.value,
        builder.potential_mapping.get_dict(),
    )
    vasp_builder.structure = builder.structure

    return dryrun_vasp(vasp_builder, **kwargs)



"""
The ``DOSCAR`` parser interface.

Contains the parsing interfaces to parsevasp used to parse ``DOSCAR`` content.
"""

# pylint: disable=abstract-method
from typing import TextIO

from parsevasp.doscar import Doscar

from aiida_vasp.parsers.content_parsers.base import BaseFileParser


class DoscarParser(BaseFileParser):
    """The parser interface that enables parsing of ``DOSCAR`` content.

    The parser is triggered by using the ``doscar-dos`` quantity key. The quantity key ``dos``
    will on the other hand parse the structure using the XML parser.

    """

    DEFAULT_SETTINGS = {'quantities_to_parse': ['doscar-dos']}

    PARSABLE_QUANTITIES = {
        'doscar-dos': {
            'inputs': [],
            'name': 'dos',
            'prerequisites': [],
        },
    }

    def _init_from_handler(self, handler: TextIO) -> None:
        """Initialize a ``parsevasp`` object of ``Doscar`` using a file like handler.

        :param handler: A file like object that provides the necessary ``DOSCAR`` content to be parsed.
        :type handler: file-like object
        """

        try:
            non_collinear = self._settings.get('non_collinear', False)
            self._content_parser = Doscar(file_handler=handler, logger=self._logger, non_collinear=non_collinear)
        except SystemExit:
            self._logger.warning('Parsevasp exited abnormally.')

    @property
    def dos(self):
        """Return the total and partial density of states, and in addition some metadata.

        :returns: A dict containing the keys ``tdos``, ``pdos`` and ``header``, which contain
                  for the two first, NumPy arrays for the total density of states and partial
                  density of states, respectively.
        :rtype: dict
        """
        metadata = self._content_parser.get_metadata()
        total_dos = self._content_parser.get_dos()
        partial_dos = self._content_parser.get_pdos()
        dos = {'tdos': total_dos, 'pdos': partial_dos, 'header': metadata}

        return dos



"""
Extensions of dictionaries.

Extensions of Pythons standard dict as well as Aiida's AttributeDict.
"""

from __future__ import annotations

import collections.abc
from collections.abc import MutableMapping  # pylint: disable=import-outside-toplevel
from contextlib import suppress  # pylint: disable=import-outside-toplevel
from copy import deepcopy
from typing import Any

from aiida import orm
from aiida.common.extendeddicts import AttributeDict


class DictWithAttributes(AttributeDict):
    """
    Extension of the AttributeDict from Aiida.common.

    This class internally stores values in a dictionary, but exposes
    the keys also as attributes, i.e. asking for attrdict.key
    will return the value of attrdict['key'] and so on.

    If the key is not in the dict a default value will be returned.
    """

    def __getattr__(self, attr: str) -> Any:
        """Read a key as an attribute. Return a Default value on missing key."""
        return self.get(attr)

    def __setattr__(self, attr: str, value: Any) -> None:
        """Set a key as an attribute."""
        self[attr] = value


def delete_keys_from_dict(dictionary: dict[str, Any], keys: str | list[str]) -> None:
    """
    Delete a key from a nested dictionary.

    Extended to support somekey.someotherkey in case we need some restrictions on the nesting.
    """
    if not isinstance(keys, list):
        keylist = [keys]
    else:
        keylist = keys
    for key in keylist:
        nested_keys = key.strip().split('.')
        delete_nested_key(dictionary, nested_keys)


def delete_nested_key(dictionary: dict[str, Any], keys: list[str]) -> None:
    """Delete the dictionary entry corresponding to a nested hierarchy of keys."""

    if keys and dictionary:
        element = keys[0]
        if element:
            value = dictionary.get(element)
            if len(keys) == 1:
                with suppress(KeyError):
                    del dictionary[element]
            elif isinstance(value, MutableMapping):
                delete_nested_key(value, keys[1:])


def update_nested_dict(dict1: dict[str, Any], dict2: dict[str, Any], extend_list: bool = False) -> dict[str, Any]:
    """Updated a nested dictionary, where dict1 is updated with values in dict2."""
    for key, value in dict2.items():
        dict1_value = dict1.get(key)
        if isinstance(value, collections.abc.Mapping) and isinstance(dict1_value, collections.abc.Mapping):
            update_nested_dict(dict1_value, value)
        elif isinstance(value, list) and isinstance(dict1_value, list) and extend_list:
            dict1_value.extend(value)
        else:
            dict1[key] = deepcopy(value)
    return dict1


def update_nested_dict_node(dict_node: orm.Dict, update_dict: dict[str, Any], extend_list: bool = False) -> orm.Dict:
    """Utility to update a Dict node in a nested way"""
    pydict = dict_node.get_dict()
    update_nested_dict(pydict, update_dict, extend_list=extend_list)
    # Check if we have updated the node. If not, return the same node.
    if pydict == dict_node.get_dict():
        return dict_node
    return orm.Dict(dict=pydict)



"""
Mock vasp code.

A more advanced way of mocking. The input of a calculation can be
hash and match from a repository of calculation that has been run.

This way we can perform tests for workchain without the need for
injecting test code into the workchain logic itself.
"""

from __future__ import annotations

import hashlib
import logging
import os
import pathlib
import shutil
import warnings
from subprocess import run

import numpy as np
from aiida import orm
from aiida.plugins import CalculationFactory
from aiida.repository import FileType
from parsevasp.incar import Incar
from parsevasp.kpoints import Kpoints
from parsevasp.poscar import Poscar

# pylint: disable=logging-format-interpolation, import-outside-toplevel

INPUT_OBJECTS = ('POSCAR', 'INCAR', 'KPOINTS')
# Objects (files) that should be excluded from storing in the repository
# These objects are mostly not used for parsing and can be large in size
DEFAULT_EXCLUDED = (
    'POTCAR',
    '.aiida',
    'LOCPOT',
    'DOSCAR',
    'PCDAT',
    'EIGENVAL',
    'OSZICAR',
    'PCDAT',
    'XDATCAR',
    'REPORT',
    'WAVECAR',
    'CHG',
)


def data_path(*args: str) -> str:
    """Return a path to a file in the test data directory."""
    path = pathlib.Path(__file__).parent.parent.parent.parent / 'tests' / 'test_data' / pathlib.Path(*args)
    path = path.resolve()
    if not path.exists():
        warnings.warn(f'Path {path} for test data does not exist.')
    return str(path)


def get_hash(dict_obj: dict | list) -> tuple[str, list[bytes]]:
    """
    Return the hash for a dictionary of arbitrary items.

    This is not meant to be robust for all cases, but should be OK for
    matching hashes of calculation inputs.

    The point here is to make the value invariant to the permutation of key orders.
    """

    # If a list is passed - convert it to a dictionary with keys being the indices
    if isinstance(dict_obj, list):
        dict_obj = dict(enumerate(dict_obj))

    rec = []
    for key_, value_ in dict_obj.items():
        key = repr(key_)
        value = value_
        # For numpy/list with floating point zero (0.0) we have to converge -0.0 to 0.0
        # as they should be equivalent
        if isinstance(value, np.ndarray):
            value[value == 0] = 0

        # Handle if value itself is float zero
        if isinstance(value, float) and value == 0:
            value = 0.0

        if isinstance(value, (dict, list)):
            rec.append(key + ':' + get_hash(value)[0])
        else:
            # Use the string representation
            rec.append(key + ':' + repr(value) + ':' + repr(type(value)))

    # Update, use sorted so the original order does not matter, in force case so
    # sting keys with upper/lower cases are treated as the same
    base = [record.encode().lower() for record in sorted(rec)]
    # Compute the hash
    md5 = hashlib.md5()
    for item in base:
        md5.update(item)

    return md5.hexdigest(), base


class MockRegistry:
    """
    A class to create and manage a registry of completed calculations.

    Calculations are identified using the hash of the parsed inputs.

    The class uses environmental variables to control its behaviour:

    - MOCK_{CODE}_REG_BASE: Prefix to the upload relative path
    - MOCK_{CODE}_UPLOAD_PREFIX: Prefix to the upload relative path

    The `{CODE}` is replaced with the `CODE_NAME` class attribute of the subclass.

    """

    CODE_NAME = 'ABSTRACT'

    def __init__(self, base_path: str | pathlib.Path | list[str | pathlib.Path] | None = None) -> None:
        """
        Instantiate and Registry
        """
        if base_path is None:
            base_path = os.environ.get(f'MOCK_{self.CODE_NAME}_REG_BASE', data_path('.'))
        base_path = pathlib.Path(base_path).absolute()

        if isinstance(base_path, (pathlib.Path, str)):
            self._search_paths = [pathlib.Path(base_path)]
        else:
            self._search_paths = [pathlib.Path(path) for path in base_path]

        self.reg_hash = {}
        self.reg_name = {}
        self.logger = logging.getLogger('aiida_vasp.utils.mock_code.MockRegistry')
        self._setup_logger()
        self.scan()

    def append_search_path(self, path: str | pathlib.Path) -> None:
        """Add a path to the list of search paths"""
        self.search_paths.append(pathlib.Path(path))

    @property
    def base_path(self) -> pathlib.Path:
        """Return the base repository path of the registry"""
        return self._search_paths[0]

    @property
    def search_paths(self) -> list[pathlib.Path]:
        """Return a list of all search paths"""
        return self._search_paths

    def _setup_logger(self, level: int = logging.INFO) -> None:
        """Setup the logger"""
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        handler = logging.StreamHandler()
        handler.setLevel(level)
        handler.setFormatter(formatter)
        self.logger.setLevel(level)

    def scan(self) -> None:
        """
        Scan the base folder and locate input/output folders
        """
        for registry_path in self.search_paths:
            for output_folder in registry_path.glob('**/out'):
                calc_base_folder = output_folder.parent.absolute()
                self._register_folder(calc_base_folder)

    def get_path_by_hash(self, hash_val: str) -> pathlib.Path:
        """
        Return the output folder for a given hash
        """
        return pathlib.Path(self.reg_hash[hash_val])

    def get_path_by_name(self, name: str) -> pathlib.Path:
        """
        Return the output folder for a given hash
        """
        return pathlib.Path(self.reg_hash[self.reg_name[name]])

    def extract_calc_by_path(
        self, rel_path: pathlib.Path | str, dst_path: pathlib.Path | str, include_inputs: bool = True
    ) -> None:
        """
        Copy the content of a give hash to a destination.

        :param rel_path: The relative path of the calculation folder to be
          extracted.
        :param dst: The destination path to be extracted to - must already exists.
        """
        rel_path = pathlib.Path(rel_path)
        dst_path = pathlib.Path(dst_path)
        found = False
        for reg_path in self.search_paths:
            base_out = reg_path / rel_path / 'out'
            base_in = reg_path / rel_path / 'inp'

            # Not a valid folder - skip this
            if not (base_out.exists() and base_in.exists()):
                continue

            found = True
            # Copy the content of input and then the output folder
            paths = [base_in, base_out] if include_inputs else [base_out]
            for folder in paths:
                for fpath in folder.glob('*'):
                    if fpath.is_file():
                        shutil.copy2(fpath, dst_path)
                    # Directory - then copy the sub files - this only handles one level down
                    elif fpath.is_dir():
                        for subfile in fpath.glob('*'):
                            shutil.copy2(subfile, dst_path / fpath.name / subfile.name)
            break
        if not found:
            raise ValueError(f'The path give: {rel_path}, is not found in any search paths.')

    def extract_calc_by_hash(self, hash_val: str, dst: pathlib.Path | str, include_inputs: bool = False) -> None:
        """
        Extract an registerred calculation using hash.
        """
        self.extract_calc_by_path(self.get_path_by_hash(hash_val), dst, include_inputs)

    def upload_calc(
        self,
        folder: pathlib.Path,
        rel_path: pathlib.Path | str,
        excluded_object: list[str] | None = None,
        included_object: list[str] | None = None,
    ) -> None:
        """
        Register a calculation folder to primary search path of the registry
        """
        inp = list(INPUT_OBJECTS)
        excluded = list(DEFAULT_EXCLUDED)
        # Exclude certain objects
        if excluded_object:
            excluded.extend(excluded_object)
        # Include certain objects
        if included_object is not None:
            for name in included_object:
                if name in excluded:
                    del excluded[excluded.index(name)]

        # Check if the repository folder already exists
        repo_calc_base = self.base_path / pathlib.Path(self.get_upload_prefix() + str(rel_path))
        if repo_calc_base.exists():
            raise FileExistsError(f'There is already a directory at {repo_calc_base.resolve()}.')

        # Deposit the objects
        repo_calc_base.mkdir(parents=True)
        repo_in = repo_calc_base / 'inp'
        repo_out = repo_calc_base / 'out'
        repo_in.mkdir(parents=True)
        repo_out.mkdir(parents=True)

        for obj in folder.glob('*'):
            if obj.name in inp:
                shutil.copy2(obj, repo_in)
            elif obj.name not in excluded:
                if obj.is_file():
                    shutil.copy2(obj, repo_out)
                elif obj.is_dir():
                    shutil.copytree(obj, repo_out / obj.name)

        # Update the hash table
        self._register_folder(repo_calc_base)

    def _register_folder(self, calc_base: pathlib.Path) -> None:
        """
        Register a folder inside the repository
        """
        # Get the relative path to the base
        rel = calc_base.relative_to(self.base_path)
        # Compute the hash
        hash_val = self.compute_hash(calc_base / 'inp')
        # Link absolute path to hash, and hash to relative path (used as name)
        self.reg_hash[hash_val] = calc_base.absolute()
        self.reg_name[str(rel)] = hash_val

    @classmethod
    def from_env(cls) -> MockRegistry:
        """Instantiate from environmental variable"""
        path = os.environ.get(f'{cls.CODE_NAME}_MOCK_CODE_BASE')
        if path is None:
            raise ValueError(f'The {cls.CODE_NAME}_MOCK_CODE_BASE environmental variable is not set!')

        paths = path.split(':')
        return cls(paths)

    @staticmethod
    def compute_hash(folder: pathlib.Path) -> str:
        """Compute the hash for a target folder"""
        raise NotImplementedError

    def upload_aiida_calc(
        self, calc_node: orm.CalcJobNode, rel_path: str | pathlib.Path, excluded_names: list[str] | None = None
    ) -> None:
        """Update a calculation into the registry"""
        raise NotImplementedError

    def upload_aiida_work(self, work_node: orm.WorkChainNode, rel_path: str | pathlib.Path) -> None:
        """Update all calculations run by an workflow into the registry"""
        raise NotImplementedError

    def get_upload_prefix(self) -> str:
        """Prefix of the name of the calculation folder"""
        prefix = os.environ.get(f'MOCK_{self.CODE_NAME}_UPLOAD_PREFIX')
        if prefix:
            prefix = prefix + '-'
        else:
            prefix = ''
        return prefix


class VaspMockRegistry(MockRegistry):
    """
    Registry of mock code for VASP

    The registry's base folder defaults to the `test_data` folder, but can be modified with the
    `MOCK_VASP_REG_BASE` environmental variable.
    """

    CODE_NAME = 'VASP'

    def upload_aiida_calc(
        self, calc_node: orm.CalcJobNode, rel_path: str | pathlib.Path, excluded_names: list[str] | None = None
    ) -> None:
        """
        Register an aiida calc_class
        """

        assert isinstance(calc_node, orm.CalcJobNode), f'{calc_node} is not an CalcJobNode!'

        # Check if the repository folder already exists
        repo_calc_base = self.base_path / pathlib.Path(self.get_upload_prefix() + str(rel_path))
        if repo_calc_base.exists():
            raise FileExistsError(f'There is already a directory at {repo_calc_base.resolve()}.')

        # Deposit the objects
        repo_calc_base.mkdir(parents=True)
        repo_in = repo_calc_base / 'inp'
        repo_out = repo_calc_base / 'out'
        repo_in.mkdir(parents=True)
        repo_out.mkdir(parents=True)

        exclude = list(DEFAULT_EXCLUDED)
        if excluded_names:
            exclude.extend(excluded_names)

        # Copy the input objects
        for obj in calc_node.base.repository.list_objects():
            if obj.name in exclude:
                continue
            copy_from_aiida(obj.name, calc_node, repo_in)

        # Copy the retrieved objects
        for obj in calc_node.outputs.retrieved.base.repository.list_objects():
            if obj.name in exclude:
                continue
            copy_from_aiida(obj.name, calc_node.outputs.retrieved, repo_out)

        self.logger.info('Calculation %s has been registered', calc_node)
        self._register_folder(repo_calc_base)

    def upload_aiida_work(self, work_node: orm.WorkChainNode, rel_path: str | pathlib.Path) -> None:
        """
        Upload all calculations in a workchain node
        """

        calc_class = CalculationFactory('vasp.vasp')
        neb_class = CalculationFactory('vasp.neb')
        to_upload = []
        for node in work_node.called_descendants:
            # Only upload VASP calculations
            if isinstance(node, orm.CalcJobNode) and (node.process_class in [calc_class, neb_class]):
                to_upload.append(node)
        to_upload.sort(key=lambda x: x.ctime)
        self.logger.info('Collected %s nodes to upload under name %s.', to_upload, rel_path)

        for idx, node in enumerate(to_upload):
            rel = pathlib.Path(rel_path) / f'calc-{idx:03d}'
            self.upload_aiida_calc(node, rel)
        self.logger.info('WorkChain %s has been uploaded.', work_node)

    @staticmethod
    def compute_hash(folder: pathlib.Path) -> str:
        """
        Compute the hash of a input folder
        """
        items = {}
        kpt_path = folder / 'KPOINTS'
        if kpt_path.is_file():
            kpoints = Kpoints(file_path=str(kpt_path))
            items['kpoints'] = kpoints.get_dict()
            items['kpoints'].pop('comment', None)

        incar_path = folder / 'INCAR'
        if incar_path.is_file():
            incar = Incar(file_path=str(incar_path), validate_tags=False)
            items['incar'] = incar.get_dict()

        poscar_path = folder / 'POSCAR'
        if not poscar_path.is_file():
            poscar_path = folder / '00/POSCAR'
        if poscar_path.is_file():
            poscar = Poscar(file_path=str(poscar_path))
            items['poscar'] = poscar.get_dict()
            items['poscar'].pop('comment', None)
        return get_hash(items)[0]


class MockVasp:
    """
    Mock VaspExecutable
    """

    def __init__(
        self,
        workdir: str | pathlib.Path,
        registry: VaspMockRegistry,
        vasp_cmd: str | list[str] | None = None,
        stdout_fname: str = 'vasp_output',
    ) -> None:
        """
        Mock VASP executable that copies over outputs from existing calculations.
        Inputs are hash and looked for.

        Notice that we do not set the hash value at init of workdir as we allow
        the unit of the MockVasp at any point, typically, you are preparing for
        a VASP calculation. Only when you execute VASP is the files checked, in this
        case when executing run. Thus, we calculate the hash of the workdir only then.

        If the `vasp_cmd` is provided the mock vasp will run the command if needed and
        upload the results to the registry. This can be useful for generating test/demo
        data.
        """
        self.workdir = workdir
        self.registry = registry
        if isinstance(vasp_cmd, str):
            vasp_cmd = [vasp_cmd]
        self.vasp_cmd = vasp_cmd
        self.stdout_fname = stdout_fname

    def run(self, debug: bool = True) -> None:
        """
        Run the mock vasp
        """

        if not os.listdir(self.workdir):
            # Directory is empty, no point of trying to find matching calcs
            raise ValueError('No input files given, so we can not find the associated test data.')

        hash_val = self.registry.compute_hash(self.workdir)
        if debug:
            print(f'Target hash value: {hash_val}')
        if hash_val in self.registry.reg_hash:
            self.registry.extract_calc_by_hash(hash_val, self.workdir)
        else:
            if debug:
                print(f'Registered hashes: {self.registry.reg_hash}')
            if self.vasp_cmd is not None:
                with open(pathlib.Path(self.workdir) / self.stdout_fname, 'w') as stdout_handle:
                    out = run(self.vasp_cmd, cwd=self.workdir, stdout=stdout_handle, check=False)
                    if out.returncode != 0:
                        raise ValueError(f'The command {self.vasp_cmd} failed with return code {out.returncode}')
                    self.registry.upload_calc(self.workdir, hash_val)
                if debug:
                    print(f'Uploaded current calculation with hash: {hash_val}')
            else:
                raise ValueError('The calculation is not registered.')

    @property
    def is_runnable(self) -> bool:
        """Check if the mock code can be executed."""
        hash_val = self.registry.compute_hash(self.workdir)
        if hash_val in self.registry.reg_hash:
            return True
        # Can we run vasp it self?
        if self.vasp_cmd is not None:
            out = run(['which', self.vasp_cmd[0]], check=False)
            if out.returncode == 0:
                return True
        return False


def copy_from_aiida(name: str, node: orm.Node, dst: pathlib.Path) -> None:
    """
    Copy objects from aiida repository.

    :param name: The full name (including the parent path) of the object.
    :param node: Node object for which the objects in the repo to be copied.
    :param dst: Path of the destination folder.

    This is a recursive function so directory copying also works.
    """
    obj = node.base.repository.get_object(name)

    # If it is a directory, copy the contents one by one
    if obj.file_type == FileType.DIRECTORY:
        for sub_obj in node.base.repository.list_objects(name):
            copy_from_aiida(os.path.join(name, sub_obj.name), node, dst)
    else:
        # Anything else
        with node.base.repository.open(name) as fsource:
            # Make parent directory if needed
            frepo_path = dst / name
            pathlib.Path(frepo_path.parent).mkdir(exist_ok=True, parents=True)
            # Write the object
            with open(frepo_path, 'w', encoding='utf8') as fdst:
                shutil.copyfileobj(fsource, fdst)



"""Test the POSCAR parser."""

# pylint: disable=unused-import,redefined-outer-name,unused-argument,unused-wildcard-import,wildcard-import, import-outside-toplevel
import numpy as np
import pytest
from parsevasp.poscar import Poscar

from aiida_vasp.parsers.content_parsers.poscar import PoscarParser, parsevasp_to_aiida


@pytest.mark.parametrize(['poscar_parser'], [('poscar',)], indirect=True)
def test_parse_poscar(poscar_parser):
    """Load a reference POSCAR parser.

    We check that it parses and provides the correct content.

    """

    # The structure for the POSCAR parser should have the key `poscar-structure`
    result = poscar_parser.get_quantity('structure')
    compare_poscar_content(result)


@pytest.mark.parametrize(['poscar_parser'], [('poscar',)], indirect=True)
def test_parse_poscar_write(poscar_parser, tmpdir):
    """Load a reference POSCAR parser and check that the write functionality works.

    Here we make sure the write function of the content parser works.

    """

    # Write the loaded structure to file
    temp_path = str(tmpdir.join('POSCAR'))
    poscar_parser.write(temp_path)

    # Load the written structure using a new content parser instance
    content = None
    with open(temp_path, 'r', encoding='utf8') as handler:
        content = handler.readlines()
    ref_content = [
        '# Compound: Al4. Old comment: Al\n',
        '  1.000000000000\n',
        '  4.040000000000   0.000000000000   0.000000000000\n',
        '  0.000000000000   4.040000000000   0.000000000000\n',
        '  0.000000000000   0.000000000000   4.040000000000\n',
        'Al\n',
        '    4\n',
        'Direct\n',
        '  0.000000000000   0.000000000000   0.000000000000\n',
        '  0.000000000000   0.500000000000   0.500000000000\n',
        '  0.500000000000   0.000000000000   0.500000000000\n',
        '  0.500000000000   0.500000000000   0.000000000000\n',
        '\n',
        '  0.000000000000   0.000000000000   0.000000000000\n',
        '  0.000000000000   0.000000000000   0.000000000000\n',
        '  0.000000000000   0.000000000000   0.000000000000\n',
        '  0.000000000000   0.000000000000   0.000000000000\n',
    ]
    assert content == ref_content


@pytest.mark.parametrize(['vasp_structure'], [('str-Al',)], indirect=True)
def test_parse_poscar_data(vasp_structure, tmpdir):
    """Load a reference AiiDA StructureData and check that the parser can
    initialize using the data.

    Using the StructureData sitting in the initialized parser it should
    write that content to a POSCAR file when issuing write which is also tested,
    file is reloaded and content checked.

    """

    # Initialize parser with an existing reference StructureData
    poscar_parser = PoscarParser(data=vasp_structure)

    # Check that get_quantity return the same StructureData instance
    assert vasp_structure == poscar_parser.get_quantity('key_does_not_matter')

    # Write the loaded StructureData to file, which behind the scenes convert it
    # to a POSCAR format
    temp_path = str(tmpdir.join('POSCAR'))
    poscar_parser.write(temp_path)

    # Load the written structure using a new content parser instance
    parser = None
    with open(temp_path, 'r', encoding='utf8') as handler:
        parser = PoscarParser(handler=handler)
    result = parser.get_quantity('structure')
    # When we start from StructureData we do not have any velocity or predictor
    # values present, so let us not compare them. The reference POSCAR has no
    # velocities, but predictors, so we can not compare those directly.
    compare_poscar_content(result, predictors=False)


@pytest.mark.parametrize(['vasp_structure'], [('str-Al',)], indirect=True)
def test_consistency_with_parsevasp(vasp_structure, data_path):
    """Compare the poscar-dict returned by parsevasp to the dict created by the PoscarParser.

    This tests purpose is to give a warning if we are overriding keys in parsevasps poscar-dict.
    """
    path = data_path('poscar', 'POSCAR')
    poscar = Poscar(file_path=path, prec=12, conserve_order=True)

    poscar_dict = poscar.get_dict(direct=False)
    result_dict = parsevasp_to_aiida(poscar)
    compare_objects(poscar_dict, result_dict)


def compare_objects(obj_a, obj_b):
    """Compare two potentially nested objects assuming they have the same structure."""
    if isinstance(obj_a, dict):
        for key in obj_a:
            compare_objects(obj_a[key], obj_b[key])
            return

    if isinstance(obj_a, (list, tuple, np.ndarray)):
        for item_a, item_b in zip(obj_a, obj_b):
            compare_objects(item_a, item_b)
            return

    assert obj_a == obj_b


def compare_poscar_content(result, velocities=True, predictors=True):
    """Compare the content of a supplied get_quantity('poscar-structure') with reference data."""
    unitcell = np.array([[4.04, 0.0, 0.0], [0.0, 4.04, 0.0], [0.0, 0.0, 4.04]])
    keys = ['comment', 'unitcell', 'sites']
    species = [
        {
            'specie': 'Al',
            'position': np.array([0.0, 0.0, 0.0]),
            'selective': [True, True, True],
            'velocities': None,
            'predictors': np.array([0.0, 0.0, 0.0]),
            'direct': False,
            'symbol': 'Al',
            'kind_name': 'Al',
        },
        {
            'specie': 'Al',
            'position': np.array([0.0, 2.02, 2.02]),
            'selective': [True, True, True],
            'velocities': None,
            'predictors': np.array([0.0, 0.0, 0.0]),
            'direct': False,
            'symbol': 'Al',
            'kind_name': 'Al',
        },
        {
            'specie': 'Al',
            'position': np.array([2.02, 0.0, 2.02]),
            'selective': [True, True, True],
            'velocities': None,
            'predictors': np.array([0.0, 0.0, 0.0]),
            'direct': False,
            'symbol': 'Al',
            'kind_name': 'Al',
        },
        {
            'specie': 'Al',
            'position': np.array([2.02, 2.02, 0.0]),
            'selective': [True, True, True],
            'velocities': None,
            'predictors': np.array([0.0, 0.0, 0.0]),
            'direct': False,
            'symbol': 'Al',
            'kind_name': 'Al',
        },
    ]
    assert set(result.keys()) == set(keys)
    assert np.allclose(result['unitcell'], unitcell)
    for index, item in enumerate(species):
        assert item['specie'] == result['sites'][index]['specie']
        assert item['selective'] == result['sites'][index]['selective']
        assert item['direct'] == result['sites'][index]['direct']
        assert item['symbol'] == result['sites'][index]['symbol']
        assert item['kind_name'] == result['sites'][index]['kind_name']
        assert np.allclose(item['position'], result['sites'][index]['position'])
        if velocities:
            assert item['velocities'] == result['sites'][index]['velocities']
        if predictors:
            assert np.allclose(item['predictors'], result['sites'][index]['predictors'])



"""
VASP Convergence WorkChain for AiiDA
====================================

This module provides a workchain for performing convergence tests in VASP calculations using the AiiDA workflow engine.
It automates the process of running multiple VASP calculations with varying plane-wave cutoff energies and
k-point spacings, collects the results, and provides utilities for analyzing and plotting the convergence data.
The pandas package is needed for generated DataFrames of the convergence test results.


Main Features
-------------

- Defines the `VaspConvergenceWorkChain`, which orchestrates a series of VASP calculations to test convergence with
  respect to:
    - Plane-wave cutoff energy (`ENCUT`)
    - K-point spacing
- Collects and summarizes output data such as total energy, maximum force, maximum stress, and magnetization.
- Provides helper functions to extract convergence data as pandas DataFrames and to plot the results.
- Includes a utility to generate a builder for the convergence workchain with user-specified protocols and options.

Inputs
------

- Structure and calculation parameters as required by the underlying VASP workchain.
- `conv_settings`: A dictionary specifying the convergence test ranges and steps for cutoff energy and k-point spacing.

Outputs
-------

- `cutoff_conv_data`: Summary of results for cutoff energy convergence.
- `kpoints_conv_data`: Summary of results for k-point spacing convergence.

Usage
-----

This module is intended to be used as part of an AiiDA plugin for VASP. Users can launch the convergence workchain
by providing the required inputs and analyze the results using the provided utility functions.


"""

from __future__ import annotations

from typing import Any

import numpy as np
from aiida import orm
from aiida.engine import ProcessSpec, WorkChain, append_, calcfunction
from aiida.plugins import WorkflowFactory

from aiida_vasp.utils.extended_dicts import update_nested_dict_node
from aiida_vasp.utils.opthold import ConvOptions

from .common.builder_updater import VaspBuilderUpdater
from .mixins import WithBuilderUpdater

# pylint:disable=no-member,unused-argument,no-self-argument,import-outside-toplevel


class VaspConvergenceWorkChain(WorkChain, WithBuilderUpdater):
    """
    A workchain to perform convergence tests.

    The inputs are essentially the same as for ``VaspWorChain`` but instead of launching
    a single calculation it launches many calculations with different kpoint spacing
    and the cut off energy.

    A ``conv_setting`` input controls the range of cut off energies and kpoint spacings.
    The available options are:

    - cutoff_start
    - cutoff_stop
    - cutoff_step
    - kspacing_start
    - kspacing_stop
    - kspacing_step
    - cutoff_kconv : cut-off energy for the kpoints convergence tests.
    - kspacing_cutconv : the kpoint spacing to be used for cut-off energy convergence tests.

    The the output data are collected and stored in two ``Dict`` output nodes.
    """

    _sub_workchain_string = 'vasp.v2.vasp'
    _sub_workchain = WorkflowFactory(_sub_workchain_string)
    ENERGY_KEY = 'energy_extrapolated'
    option_class = ConvOptions

    @classmethod
    def define(cls, spec: ProcessSpec) -> None:
        super().define(spec)

        spec.expose_inputs(cls._sub_workchain)
        spec.input(
            'conv_settings',
            help=ConvOptions.aiida_description(),
            validator=ConvOptions.aiida_validate,
            serializer=ConvOptions.aiida_serialize,
            valid_type=orm.Dict,
        )
        spec.outline(cls.setup, cls.launch_conv_calcs, cls.analyse)

        spec.exit_code(
            401,
            'ERROR_SUBWORKFLOW_ERRORED',
            message='At leaste one of the launched sub-workchain has failed',
        )
        spec.output('kpoints_conv_data', required=False)
        spec.output('cutoff_conv_data', required=False)

    def setup(self) -> None:
        """Setup the convergence workflow"""
        settings = self.inputs.conv_settings.get_dict()
        self.ctx.settings = settings

        # Planewave cut off energies
        start = settings['cutoff_start']
        stop = settings['cutoff_stop']
        if start < stop:
            cutoff_list = [start]
            cut = start
            # Ensure start and stop are always included
            while True:
                cut += settings['cutoff_step']
                if cut < stop:
                    cutoff_list.append(cut)
                else:
                    cutoff_list.append(stop)
                    break
        else:
            # Start is equal or larger than stop - signalling no need to do the test
            cutoff_list = []

        # Same treatment for kspacing
        start = settings['kspacing_start']
        stop = settings['kspacing_stop']

        if start > stop:
            spacing = start
            kspacing_list = [spacing]
            while True:
                spacing -= settings['kspacing_step']
                if spacing > settings['kspacing_stop']:
                    kspacing_list.append(spacing)
                else:
                    kspacing_list.append(settings['kspacing_stop'])
                    break
        else:
            kspacing_list = []

        self.ctx.cutoff_list = cutoff_list
        self.ctx.kspacing_list = kspacing_list

    def launch_conv_calcs(self) -> None:
        """
        Setup and launch the convergence calculations
        """
        if not self.ctx.cutoff_list:
            cut_k = 400  # Default if not supplied
        else:
            cut_k = min(self.ctx.cutoff_list)
        if not self.ctx.kspacing_list:
            k_cut = 0.06  # Default if not supplied
        else:
            k_cut = min(self.ctx.kspacing_list)

        cutoff_for_kconv = self.ctx.settings.get('cutoff_kconv', cut_k)
        kspacing_for_cutoffconv = orm.Float(self.ctx.settings.get('kspacing_cutconv', k_cut))

        # Launch cut off energy tests
        inputs = self.exposed_inputs(self._sub_workchain)
        inputs.kpoints_spacing = kspacing_for_cutoffconv
        original_label = inputs.metadata.get('label', '')
        for cut in self.ctx.cutoff_list:
            new_param = update_nested_dict_node(inputs.parameters, {'incar': {'encut': cut}})
            inputs.parameters = new_param
            if original_label:
                inputs.metadata.label = original_label + f' CUTCONV {cut:.2f}'
            else:
                inputs.metadata.label = f'CUTCONV {cut:.2f}'

            running = self.submit(self._sub_workchain, **inputs)
            self.report(f'Submitted {running} with cut off energy {cut:.1f} eV.')
            self.to_context(cutoff_conv_workchains=append_(running))

        # Launch kpoints convergence tests
        new_param = update_nested_dict_node(inputs.parameters, {'incar': {'encut': cutoff_for_kconv}})
        for kspacing in self.ctx.kspacing_list:
            inputs.parameters = new_param
            inputs.kpoints_spacing = kspacing
            if original_label:
                inputs.metadata.label = original_label + f' KCONV {kspacing:.3f}'
            else:
                inputs.metadata.label = f'KCONV {kspacing:.3f}'

            running = self.submit(self._sub_workchain, **inputs)
            self.report(f'Submitted {running} with kpoints spacing {kspacing:.3f}.')
            self.to_context(kpoints_conv_workchains=append_(running))

    def analyse(self) -> None:
        """
        Analyse the output of the calculations.
        Collect data to be plotted/analysed against the cut off energy and kpoints spacing
        """

        def get_maximum(forces: np.ndarray | None) -> float | None:
            if forces is None:
                return None
            norm = np.linalg.norm(forces, axis=1)
            return np.amax(norm)

        def collect_data(workchain: orm.WorkChainNode, energy_key: str) -> dict[str, Any]:
            """Collect the data from workchain output"""
            output = workchain.outputs.misc.get_dict()
            data = {}
            data['maximum_force'] = get_maximum(output.get('forces'))
            # Extract the magnetization
            magnetization = output.get('magnetization')
            if magnetization:
                data['magnetization'] = magnetization[0]
            data['maximum_stress'] = get_maximum(output.get('stress'))
            data['energy'] = output['total_energies'][energy_key]
            return data

        def unpack(name: str, input_data: dict[Any, Any]) -> dict[str, list[Any]]:
            """Unpack a dict with numberical keys"""
            output_dict = {name: []}
            for key, data in input_data.items():
                output_dict[name].append(key)
                # Append values to the corresponding lists
                for key_, value in data.items():
                    if key_ not in output_dict:
                        output_dict[key_] = []
                    output_dict[key_].append(value)
            return output_dict

        exit_code = None

        cutoff_data = {}
        cutoff_miscs = {}
        energy_key = None
        if 'cutoff_conv_workchains' in self.ctx:
            for iwork, workchain in enumerate(self.ctx.cutoff_conv_workchains):
                if workchain.exit_status != 0:
                    exit_code = self.exit_codes.ERROR_SUBWORKFLOW_ERRORED
                    self.report(f'Skipping workchain {workchain} with exit status {workchain.exit_status} ')
                    continue

                # Setup the energy key from the first workchain
                if not energy_key:
                    energy_key = next(iter(workchain.outputs.misc.get_dict()['total_energies'].keys()))

                cutoff = workchain.inputs.parameters['incar']['encut']
                cutoff_data[cutoff] = collect_data(workchain, energy_key)
                cutoff_data[cutoff]['mesh'] = workchain.called[0].inputs.kpoints.get_kpoints_mesh()[0]
                cutoff_miscs[f'worchain_{iwork}'] = workchain.outputs.misc

        kspacing_data = {}
        kspacing_miscs = {}

        if 'kpoints_conv_workchains' in self.ctx:
            for iwork, workchain in enumerate(self.ctx.kpoints_conv_workchains):
                if workchain.exit_status != 0:
                    exit_code = self.exit_codes.ERROR_SUBWORKFLOW_ERRORED
                    self.report(f'Skipping Workchain {workchain} with exit status {workchain.exit_status} ')
                    continue

                # Setup the energy key from the first workchain
                if not energy_key:
                    energy_key = next(iter(workchain.outputs.misc.get_dict()['total_energies'].values()))

                spacing = float(workchain.inputs.kpoints_spacing)
                kspacing_data[spacing] = collect_data(workchain, energy_key)
                kspacing_data[spacing]['mesh'] = workchain.called[0].inputs.kpoints.get_kpoints_mesh()[0]
                kspacing_miscs[f'worchain_{iwork}'] = workchain.outputs.misc
                cutoff = workchain.inputs.parameters['incar']['encut']
                kspacing_data[spacing]['cutoff_energy'] = cutoff

        # Calcfunction to link with the calculation output to the summary data node
        @calcfunction
        def create_links_kconv(**miscs: Any) -> orm.Dict:
            """Alias calcfunction to link summary node with miscs"""
            return orm.Dict(dict=unpack('kpoints_spacing', kspacing_data))

        @calcfunction
        def create_links_cutconv(**miscs: Any) -> orm.Dict:
            """Alias calcfunction to link summary node with miscs"""
            return orm.Dict(dict=unpack('cutoff_energy', cutoff_data))

        if kspacing_data:
            self.out('kpoints_conv_data', create_links_kconv(**kspacing_miscs))
        if cutoff_data:
            self.out('cutoff_conv_data', create_links_cutconv(**cutoff_miscs))

        return exit_code

    @staticmethod
    def get_conv_data(conv_work: orm.WorkChainNode, plot: bool = False, **plot_kwargs: Any) -> tuple[Any, Any]:
        """
        Convenient method for extracting convergence data

        Args:
        conv_work (orm.WorkChainNode): Convergence workflow node

        Returns:
        A tuple of cut-off convergence and k-point convergence result dataframe
        """
        cdf, kdf = get_conv_data(conv_work)
        if plot is True:
            plot_conv_data(cdf, kdf, **plot_kwargs)
        return cdf, kdf


def get_conv_data(conv_work: orm.WorkChainNode) -> tuple[Any, Any]:
    """
    Convenient method for extracting convergence data

    Args:
      conv_work (orm.WorkChainNode): Convergence workflow node

    Returns:
      A tuple of cut-off convergence and k-point convergence result data frame
    """
    import pandas as pd  # noqa: PLC0415

    if 'cutoff_conv_data' in conv_work.outputs:
        cutdf = pd.DataFrame(conv_work.outputs.cutoff_conv_data.get_dict())
        cutdf['energy_per_atom'] = cutdf['energy'] / len(conv_work.inputs.structure.sites)
        cutdf['dE_per_atom'] = cutdf['energy_per_atom'] - cutdf['energy_per_atom'].iloc[-1]
    else:
        cutdf = None

    if 'kpoints_conv_data' in conv_work.outputs:
        kdf = pd.DataFrame(conv_work.outputs.kpoints_conv_data.get_dict())
        kdf['energy_per_atom'] = kdf['energy'] / len(conv_work.inputs.structure.sites)
        kdf['dE_per_atom'] = kdf['energy_per_atom'] - kdf['energy_per_atom'].iloc[-1]
    else:
        kdf = None
    return cutdf, kdf


def plot_conv_data(cdf: Any, kdf: Any, **kwargs: Any) -> list[Any]:
    """
    Make two combined plots for the convergence test results.
    """

    import matplotlib.pyplot as plt  # noqa: PLC0415

    # Create a subplot
    figs = []
    if cdf is not None:
        fig, axs = plt.subplots(3, 1, sharex=True, **kwargs)
        figs.append(fig)
        axs[0].plot(cdf.cutoff_energy, cdf.dE_per_atom, '-x')
        axs[0].set_ylabel('dE (eV / atom)')
        i = 0
        if 'maximum_force' in cdf.columns:
            i += 1
            axs[i].plot(cdf.cutoff_energy, cdf.maximum_force, '-x')
            axs[i].set_ylabel(r'$F_{max}$ (eV$\AA^{-1}$)')
        if 'maximum_stress' in cdf.columns:
            i += 1
            axs[i].plot(cdf.cutoff_energy, cdf.maximum_stress, '-x')
            axs[i].set_ylabel(r'$S_{max}$ (kBar)')
        axs[i].set_xlabel('Cut-off energy (eV)')
        fig.tight_layout()

    if kdf is not None:
        fig, axs = plt.subplots(3, 1, sharex=True, **kwargs)
        figs.append(fig)
        axs[0].plot(kdf.kpoints_spacing, kdf.dE_per_atom, '-x')
        axs[0].set_ylabel('dE (eV / atom)')
        i = 0
        if 'maximum_force' in kdf.columns:
            i += 1
            axs[i].plot(kdf.kpoints_spacing, kdf.maximum_force, '-x')
            axs[i].set_ylabel(r'$F_{max}$ (eV$\AA^{-1}$)')
        if 'maximum_stress' in kdf.columns:
            i += 1
            axs[i].plot(kdf.kpoints_spacing, kdf.maximum_stress, '-x')
            axs[i].set_ylabel(r'$S_{max}$ (kBar)')
        axs[i].set_xticks(kdf.kpoints_spacing)
        axs[i].set_xticklabels(
            [f'{row.kpoints_spacing:.3f}\n{row.mesh}' for _, row in kdf.iterrows()],
            rotation=45,
        )
        axs[i].set_xlabel('K-pointing spacing (mesh)')
        fig.tight_layout()

    return figs


def get_convergence_builder(structure: orm.StructureData, config: dict[str, Any]) -> VaspBuilderUpdater:
    """
    Short cut for getting an VaspBuilderUpdater ready to use

    :structure StructureData: The input structure node.
    :config dict: Configuration dictionary specifying the protocol.

    The following files are used from the configuration: ``code``, ``inputset``, ``conv``, ``options``, ``resources``.
    """

    conv_builder = VaspConvergenceWorkChain.get_builder()

    upd = VaspBuilderUpdater(conv_builder)
    upd.use_inputset(
        structure,
        config.get('inputset', VaspBuilderUpdater.DEFAULT_INPUTSET),
        overrides=config.get('overrides', {}),
    )
    upd.set_code(orm.load_code(config['code']))

    upd.set_default_options(**config.get('options', {}))
    upd.update_resources(**config.get('resources', {}))
    upd.set_label(f'{structure.label} CONV')

    # Convergence specific options
    conv = ConvOptions(**config.get('conv', {}))
    upd.builder.conv_settings = conv.aiida_dict()
    return upd



"""Unit tests for VASP calculation monitoring functions."""

import os
import time
from pathlib import Path

import pytest
from aiida.common.extendeddicts import AttributeDict

from aiida_vasp.calcs.monitors import monitor_loop_time, monitor_stdout
from aiida_vasp.calcs.vasp import VaspCalculation


class MockFileStatResult:
    """Mock object for file stat results."""

    def __init__(self, size=1024, mtime=None):
        """Initialize mock file stat result.

        :param size: File size in bytes
        :param mtime: File modification time (defaults to current time)
        """
        self.st_size = size
        self.st_mtime = mtime if mtime is not None else time.time()


class MockTransport:
    """Mock transport object for testing."""

    def __init__(self):
        """Initialize mock transport."""
        self.get_attribute_return = None
        self.exec_command_wait_return = None
        self.get_attribute_calls = []
        self.exec_command_wait_calls = []

    def reset(self):
        """Reset the status"""
        self.get_attribute_return = None
        self.exec_command_wait_return = None
        self.get_attribute_calls = []
        self.exec_command_wait_calls = []

    def get_attribute(self, filename):
        """Mock get_attribute method."""
        self.get_attribute_calls.append(filename)
        return self.get_attribute_return

    def exec_command_wait(self, command):
        """Mock exec_command_wait method."""
        self.exec_command_wait_calls.append(command)
        if self.exec_command_wait_return is None:
            return (0, '', '')
        return self.exec_command_wait_return


class MockProcessClass:
    """Mock process class with VASP output name."""

    def __init__(self, vasp_output='vasp_output'):
        """Initialize with VASP output filename."""
        self._VASP_OUTPUT = vasp_output


class MockNode:
    """Mock CalcJobNode for testing."""

    def __init__(self, walltime_limit=None, vasp_output='vasp_output'):
        """Initialize mock node."""
        self.process_class = MockProcessClass(vasp_output)
        self.inputs = AttributeDict()
        self.options = {'max_wallclock_seconds': walltime_limit}

    def get_option(self, key):
        """Mocked get_option method"""
        return self.options[key]

    def get_remote_workdir(self):
        """Mocked get_remote_workdir method"""
        return os.getcwd()  # Use current working directory for simplicity


@pytest.fixture
def mock_transport():
    """Provide a mock transport object."""
    return MockTransport()


@pytest.fixture
def mock_node():
    """Provide a mock CalcJobNode."""
    return MockNode()


class TestMonitorStdout:
    """Test cases for monitor_stdout function."""

    def test_monitor_stdout_normal_size(self, mock_node, mock_transport):
        """Test monitor_stdout with normal file size."""
        # Setup mocks
        mock_transport.get_attribute_return = MockFileStatResult(size=1024 * 1024)  # 1MB

        # Call function
        result = monitor_stdout(mock_node, mock_transport, size_threshold_mb=5)

        # Assertions
        stdout_path = str(Path(os.getcwd()) / VaspCalculation._VASP_OUTPUT)
        assert result is None
        assert mock_transport.get_attribute_calls == [stdout_path]
        assert mock_transport.exec_command_wait_calls == []

    def test_monitor_stdout_oversized_file(self, mock_node, mock_transport):
        """Test monitor_stdout with oversized file triggering truncation."""
        # Setup mocks
        file_size = 10 * 1024 * 1024  # 10MB
        mock_transport.get_attribute_return = MockFileStatResult(size=file_size)

        # Call function
        result = monitor_stdout(mock_node, mock_transport, size_threshold_mb=5)

        # Assertions
        assert result is not None
        assert 'Very large stdout detected' in result
        assert '10.00 MB' in result
        assert 'potential critical crash' in result

        stdout_path = str(Path(os.getcwd()) / VaspCalculation._VASP_OUTPUT)
        assert mock_transport.get_attribute_calls == [stdout_path]
        assert mock_transport.exec_command_wait_calls == [f'truncate -s 5M {stdout_path}']

        # Apply a custom threshold
        mock_transport.reset()
        mock_transport.get_attribute_return = MockFileStatResult(size=file_size)
        result = monitor_stdout(mock_node, mock_transport, size_threshold_mb=15)

        assert result is None
        assert not mock_transport.exec_command_wait_calls

        # Also pass if exactly at threshold
        # Apply a custom threshold
        mock_transport.reset()
        mock_transport.get_attribute_return = MockFileStatResult(size=file_size)
        result = monitor_stdout(mock_node, mock_transport, size_threshold_mb=10)
        assert result is None
        assert not mock_transport.exec_command_wait_calls


class TestMonitorLoopTime:
    """Test cases for monitor_loop_time function."""

    def test_monitor_loop_time_no_walltime_limit(self, mock_transport):
        """Test monitor_loop_time when no walltime limit is set."""
        # Setup node without walltime limit
        mock_node = MockNode(walltime_limit=None)

        # Call function
        result = monitor_loop_time(mock_node, mock_transport)

        # Assertions
        assert result is None
        assert mock_transport.exec_command_wait_calls == []

    def test_monitor_loop_time_no_loop_entries(self, mock_transport):
        """Test monitor_loop_time when no LOOP entries are found in OUTCAR."""
        # Setup node with walltime
        mock_node = MockNode(walltime_limit=3600)  # 1 hour
        mock_transport.exec_command_wait_return = (1, '', 'grep: no match')

        # Call function
        result = monitor_loop_time(mock_node, mock_transport)

        # Assertions
        assert result is None
        outcar_path = Path(os.getcwd()) / 'OURCAR'
        assert mock_transport.exec_command_wait_calls == [f"grep 'LOOP:' {outcar_path}"]

    def test_monitor_loop_time_fast_loops(self, mock_transport):
        """Test monitor_loop_time with fast electronic loops."""
        # Setup node with walltime
        mock_node = MockNode(walltime_limit=3600)  # 1 hour

        # Mock grep output with LOOP entries (fast loops - 10 seconds each)
        # Format: "LOOP:  CPU time  real_time"
        grep_output = 'LOOP:  cpu time    4.0: real time    4.0\nLOOP:  cpu time    4.0: real time    4.0'
        mock_transport.exec_command_wait_return = (0, grep_output, '')
        mock_transport.get_attribute_return = MockFileStatResult(mtime=time.time())

        # Call function
        result = monitor_loop_time(mock_node, mock_transport)

        # Assertions - should be fine as 12.3s < 3600s/10 = 360s
        assert result is None

    def test_monitor_loop_time_slow_loops(self, mock_transport):
        """Test monitor_loop_time with slow electronic loops."""
        # Setup node with walltime
        mock_node = MockNode(walltime_limit=3600)  # 1 hour

        # Mock grep output with LOOP entries (slow loops - 400 seconds each)
        grep_output = 'LOOP:  cpu time    400.0: real time    400\nLOOP:  cpu time    400.0: real time    400.0'
        mock_transport.exec_command_wait_return = (0, grep_output, '')
        mock_transport.get_attribute_return = MockFileStatResult(mtime=time.time())

        # Call function (default minimum_electronic_loops=10)
        result = monitor_loop_time(mock_node, mock_transport)

        # Assertions - should detect slow loops as 450.3s > 3600s/10 = 360s
        assert result is not None
        assert 'Less than 10 electronic loop can be run' in result
        assert '400.00 seconds' in result

    def test_monitor_loop_time_custom_minimum_loops(self, mock_transport):
        """Test monitor_loop_time with custom minimum electronic loops."""
        # Setup node with walltime
        mock_node = MockNode(walltime_limit=3600)  # 1 hour

        # Mock grep output with LOOP entries (200 seconds each)
        grep_output = 'LOOP:  cpu time    400.0: real time    400\nLOOP:  cpu time    400.0: real time    400.0'
        mock_transport.exec_command_wait_return = (0, grep_output, '')
        mock_transport.get_attribute_return = MockFileStatResult(mtime=time.time())

        # Call function with minimum_electronic_loops=20
        result = monitor_loop_time(mock_node, mock_transport, minimum_electronic_loops=20)

        assert result is not None
        assert 'Less than 20 electronic loop can be run' in result
        assert '400.00 seconds' in result

    def test_monitor_loop_time_stalled_calculation(self, mock_transport, monkeypatch):
        """Test monitor_loop_time detecting stalled calculation."""
        # Setup time mock - current time is much later than file modification time
        current_time = 10000
        file_mtime = 8000  # File last modified 2000 seconds ago
        monkeypatch.setattr(time, 'time', lambda: current_time)

        # Setup node with walltime
        mock_node = MockNode(walltime_limit=7200)  # 2 hour

        # Mock grep output with LOOP entries (100 seconds each loop)
        grep_output = 'LOOP:  cpu time    100.0: real time    100\nLOOP:  cpu time    100.0: real time    100.0'
        mock_transport.exec_command_wait_return = (0, grep_output, '')
        mock_transport.get_attribute_return = MockFileStatResult(mtime=file_mtime)

        # Call function (default patience_num_loops=5, patience_minimum_time=1800)
        result = monitor_loop_time(mock_node, mock_transport)

        assert result is not None
        assert 'file is more than 500.00 seconds ago' in result

    def test_monitor_loop_time_recent_update(self, mock_transport, monkeypatch):
        """Test monitor_loop_time with recent file update."""
        # Setup time mock - file was modified recently
        current_time = 10000
        file_mtime = 9900  # File last modified 100 seconds ago
        monkeypatch.setattr(time, 'time', lambda: current_time)

        # Setup node with walltime
        mock_node = MockNode(walltime_limit=7200)  # 2 hour

        # Mock grep output with LOOP entries (50 seconds each loop)
        grep_output = 'LOOP:  cpu time    50.5: real time    50.5\nLOOP:  cpu time    50.5: real time    50.5'
        mock_transport.exec_command_wait_return = (0, grep_output, '')
        mock_transport.get_attribute_return = MockFileStatResult(mtime=file_mtime)

        # Call function
        result = monitor_loop_time(mock_node, mock_transport)

        # Assertions
        # elapsed = 10000 - 9900 = 100s
        # last_loop_time * patience_num_loops = 50.5 * 5 = 252.5s > 100s (not stalled)
        assert result is None

    def test_monitor_loop_time_minimum_patience_not_met(self, mock_transport, monkeypatch):
        """Test monitor_loop_time when minimum patience time is not met."""
        # Setup time mock
        current_time = 10000
        file_mtime = 8500  # File last modified 1500 seconds ago
        monkeypatch.setattr(time, 'time', lambda: current_time)

        # Setup node with walltime
        mock_node = MockNode(walltime_limit=7200)  # 2 hour

        # Mock grep output with LOOP entries
        grep_output = 'LOOP:  cpu time    100.5: real time    100.5\nLOOP:  cpu time    100.5: real time    100.5'
        mock_transport.exec_command_wait_return = (0, grep_output, '')
        mock_transport.get_attribute_return = MockFileStatResult(mtime=file_mtime)

        # Call function
        result = monitor_loop_time(mock_node, mock_transport)

        assert result is None



"""
Utils for bands structures.

Utilities for working with band structures. Currently this is legacy and will be
rewritten or moved.
"""

# ruff: noqa: PLC0415

from __future__ import annotations

import itertools
from typing import Any

import numpy as np
from aiida import orm


def get_bs_dims(bands_array: np.ndarray) -> tuple[int, int, int]:
    """
    Get the dimensions from the bands array of a BandsData node.

    :param numpy.array bands_array:
        an array with bands as stored in an array.bands data node
    :return: a tuple containing num_bands, num_kp, num_spins.
        if the array is only 2d, num_spins = 0
    :rtype tuple:
    """
    bshape = bands_array.shape
    nbd = nkp = nsp = 0
    if len(bshape) == 2:
        nbd = bshape[1]
        nkp = bshape[0]
    elif len(bshape) == 3:
        nbd = bshape[2]
        nkp = bshape[1]
        nsp = bshape[0]
    return nbd, nkp, nsp


def get_kp_labels(
    bands_node: orm.BandsData, kpoints_node: orm.KpointsData | None = None
) -> tuple[list[int], list[str]]:
    """
    Get Kpoint labels with their x-positions in matplotlib compatible format.

    A KpointsData node can optionally be given to fall back to if no labels
    are found on the BandsData node. The caller is responsible for ensuring
    the nodes match.  This should be the case if you take the kpoints from
    the input and the bands from the
    output of a calculation node.

    :param BandsData bands_node:
        The BandsData node will be searched labels first
    :param KpointsData kpoints_node:
        The optional KpointsData node will be searched only if no labels are
        present on the BandsData node. No consistency checks are performed.
    :return:  (kpx, kpl), the x-coordinates and text labels
    :rtype: tuple(list[int], list[unicode])
    :raises AttributeError: if neither of the given nodes have a labels
        attribute
    """
    kplabs = None
    kpx = []
    kpl = []
    try:
        kplabs = bands_node.labels
    except AttributeError as err:
        if kpoints_node:
            kplabs = kpoints_node.labels
        else:
            raise err
    if kplabs:
        kpx = [i[0] for i in kplabs]
        kpl = [i[1] for i in kplabs]
        for i, kpoints in enumerate(kpl):
            if kpoints == 'G':
                kpl[i] = r'$\Gamma$'
    return kpx, kpl


def get_efermi(calc: orm.CalcJobNode) -> float | None:
    """Get the fermi energy from a finished calculation."""
    efermi = None
    if calc:
        p_res = calc.get_outputs_dict().get('results')
        efermi = p_res and p_res.get_dict().get('efermi')
    return efermi


def get_kp_node(calc: orm.CalcJobNode) -> orm.KpointsData | None:
    kpoints_node = None
    if calc:
        kpoints_node = calc.get_inputs_dict().get('kpoints')
    return kpoints_node


def plot_bstr(
    bands_node: orm.BandsData,
    kpoints_node: orm.KpointsData | None = None,
    title: str | None = None,
    efermi: float | None = None,
    use_parent_calc: bool = False,
    **kwargs: Any,
):
    """
    Use matplotlib to plot the bands stored in a BandsData node.

    A KpointsData node can optionally be given as a fallback for
    kpoint labels. The caller is responsible for giving a node
    with matching labels (as in they are in/out nodes of the same
    calculation).

    :param BandsData bands_node:
        The BandsData node will be searched labels first
    :param KpointsData kpoints_node:
        The optional KpointsData node will be searched only if no labels are
        present on the BandsData node. No consistency checks are performed.
    :return: the matplotlib figure containing the plot
    """
    import matplotlib.pyplot as plt

    fig = plt.figure()
    title = title or f'Band Structure (pk={bands_node.pk})'
    bands = bands_node.get_bands()
    _, nkp, _ = get_bs_dims(bands)
    plot_bands(bands_node, **kwargs)

    parent_calc = None
    if use_parent_calc:
        inputs = bands_node.get_inputs()
        parent_calc = inputs[0] if inputs else None

    efermi = get_efermi(parent_calc)
    kpoints_node = get_kp_node(parent_calc)

    if efermi:
        plt.hlines(efermi, plt.xlim()[0], nkp - 1, linestyles='dashed')
        plt.yticks(list(plt.yticks()[0]) + [efermi], [str(line) for line in plt.yticks()[0]] + [r'$E_{fermi}$'])

    try:
        kpx, kpl = get_kp_labels(bands_node, kpoints_node)
        plt.xticks(kpx, kpl)
        plt.vlines(kpx, plt.ylim()[0], plt.ylim()[1])
    except Exception:  # pylint: disable=broad-except
        pass

    plt.ylabel('Dispersion')
    plt.suptitle(title)
    return fig


def plot_bands(bands_node: orm.BandsData, **kwargs: Any) -> None:
    """Plot a bandstructure node using matplotlib."""

    import matplotlib.pyplot as plt

    bands = bands_node.get_bands()
    nbands, nkp, nspin = get_bs_dims(bands)
    if nspin > 0:
        allbands = np.empty((nkp, nbands * nspin))
        for i in range(nspin):
            allbands[:, i * nbands : (i + 1) * nbands] = bands[i]
        bands = allbands

    if 'colors' in kwargs:
        colors = itertools.cycle(kwargs.pop('colors'))
        for b_idx in range(bands.shape[1]):
            plt.plot(bands[:, b_idx], color=colors.next(), **kwargs)
    else:
        plt.plot(bands, **kwargs)



"""
The ``OUTCAR`` parser interface.

Contains the parsing interfaces to parsevasp used to parse ``OUTCAR`` content.
"""

# pylint: disable=abstract-method
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from parsevasp.outcar import Outcar

from aiida_vasp.parsers.content_parsers.base import BaseFileParser


class OutcarParser(BaseFileParser):
    """The parser interface that enables parsing of ``OUTCAR`` content.

    The parser is triggered by using the ``elastic_moduli``, ``magnetization`` or ``site-magnetization``
    ``run_stats`` or ``run_status`` quantity keys.

    """

    DEFAULT_SETTINGS = {'quantities_to_parse': ['run_status', 'run_stats']}

    PARSABLE_QUANTITIES = {
        'elastic_moduli': {'inputs': [], 'name': 'elastic_moduli', 'prerequisites': []},
        'symmetries': {'inputs': [], 'name': 'symmetries', 'prerequisites': []},
        'magnetization': {'inputs': [], 'name': 'magnetization', 'prerequisites': []},
        'site_magnetization': {'inputs': [], 'name': 'site_magnetization', 'prerequisites': []},
        'run_stats': {
            'inputs': [],
            'name': 'run_stats',
            'prerequisites': [],
        },
        'run_status': {
            'inputs': [],
            'name': 'run_status',
            'prerequisites': [],
        },
    }

    def _init_from_handler(self, handler: Any) -> None:
        """Initialize a ``parsevasp`` object of ``Outcar`` using a file like handler.

        :param handler: A file like object that provides the necessary ``OUTCAR`` content to be parsed.
        :type handler: file-like object
        """

        try:
            self._content_parser = Outcar(file_handler=handler, logger=self._logger)
        except SystemExit:
            self._logger.warning('Parsevasp exited abnormally.')

    @property
    def run_status(self):
        """Fetch status of calculations.

        :returns: A dictionary containing the keys ``finished``, which is True if the VASP calculation
                  contain timing information in the end of the ``OUTCAR``. The key ``ionic_converged`` is
                  True if the number of ionic steps detected is smaller than the supplied NSW.
                  The key ``electronic_converged`` is True if the number of electronic steps is smaller than
                  NELM (defaults to 60 in VASP). It is also possible to check if all the ionic steps
                  did reached NELM and thus did not converged if the key ``consistent_nelm_breach`` is ``True``,
                  while ``contains_nelm_breach`` is True if one or more ionic steps reached NELM and thus
                  did not converge electronically.
        :rtype: dict
        """
        status = self._content_parser.get_run_status()
        return status

    @property
    def run_stats(self):
        """Fetch the run statistics, which included timings and memory consumption.

        :returns: A dictionary containing timing and memory consumption information
                  that are parsed from the end of the ``OUTCAR`` file. The key names are
                  mostly preserved, except for the memory which is prefixed with ``mem_usage_``.
                  Units are preserved from ``OUTCAR`` and there are some differences between
                  VASP 5 and 6.
        :rtype: dict
        """
        stats = self._content_parser.get_run_stats()
        return stats

    @property
    def symmetries(self):
        """Fetch some basic symmetry data.

        :returns: A dictionary containing the number of space group operations in the
                  key ``num_space_group_operations`` and the detected supplied cell in
                  ``original_cell_type``. In ``symmetrized_cell_type`` the cell on which
                  VASP performs the calculation has been included. Each value in the
                  dictionary is a list, where each entry represent one ionic step.
        :rtype: dict
        """

        sym = self._content_parser.get_symmetry()
        return sym

    @property
    def elastic_moduli(self):
        """Fetch the elastic moduli tensor.

        :returns: A dictionary containing ndarrays with the rigid ion elastic moduli, both symmetrized and
                  non-symmetrized for the keys ``symmetrized`` and ``non_symmetrized`` respectively.
                  The key ``total`` contain both the rigid ion and the ionic contributions to the
                  elastic tensor for the symmetrized case.
        :rtype: dict
        """

        moduli = self._content_parser.get_elastic_moduli()
        return moduli

    @property
    def site_magnetization(self):
        """Fetch the site dependent magnetization.

        :returns: A dictionary containing the key ``sphere`` which contains the integrated
                  magnetization in units of Bohr magneton. Additional keys under ``sphere`` are
                  given for each direction and for non-collinear calculations all of them are used.
                  The ``site_moment`` yields the magnetization per site, with a key describing the
                  site number and then the ``s``, ``p``, ``d`` etc. the projections of the site magnetization
                  and ``tot`` containing the total magnetization for that site.
                  The ``total_magnetization`` gives the sum of each magnetization projection and
                  magnetization total for each site.
                  The ``full_cell`` key yields the magnetization from the electronic part of the last
                  electronic step in a list.
        :rtype: dict
        """
        magnetization = self._content_parser.get_magnetization()
        return magnetization

    @property
    def magnetization(self):
        """Fetch the full cell magnetization.

        :returns: A list containing an entry that is the total magnetization in the cell in unit of
                  Bohr magneton. The magnetization returned is the one associated with the electrons for the
                  last electronic step.
        :rtype: list
        """
        magnetization = self.site_magnetization
        if magnetization is not None:
            magnetization = magnetization['full_cell']
        return magnetization


class VtstNebOutcarParser(OutcarParser):
    """
    Parser for processing OUTCAR generated by VASP with VTST
    """

    DEFAULT_SETTINGS = {'quantities_to_parse': ['run_status', 'run_stats', 'neb_data']}
    PARSABLE_QUANTITIES = {
        'neb_data': {'inputs': [], 'name': 'neb_data', 'prerequisites': []},
        'forces': {'inputs': [], 'name': 'forces', 'prerequisites': []},
        'outcar_positions': {'inputs': [], 'name': 'positions', 'prerequisites': []},
        'outcar_cell': {'inputs': [], 'name': 'cell', 'prerequisites': []},
        'total_energies': {'inputs': [], 'name': 'total_energies', 'prerequisites': []},
        **OutcarParser.PARSABLE_QUANTITIES,
    }

    def __init__(self, *args, **kwargs):
        """Instantiate the parser."""
        self._parsed_neb_data = {}
        super().__init__(*args, **kwargs)

    def _init_from_handler(self, handler: Any) -> None:
        """Initial from the handler."""
        super()._init_from_handler(handler)
        # Parse the NEB results from the handle and store in a dictionary
        self._parsed_neb_data = _parse_neb_outputs(handler)

    @property
    def neb_data(self):
        """
        Parsed NEB results
        """
        return self._parsed_neb_data.get('neb_data')

    @property
    def forces(self):
        """Parsed forces"""
        return self._parsed_neb_data.get('forces')

    @property
    def outcar_positions(self):
        """Parsed positions"""
        return self._parsed_neb_data.get('outcar_positions')

    @property
    def outcar_cell(self):
        """Parsed cell vectors"""
        return self._parsed_neb_data.get('outcar_cell')

    @property
    def total_energies(self):
        """Parsed total energies"""
        return self._parsed_neb_data.get('total_energies')


def _parse_force_block(lines: List[str]) -> Tuple[List[List[float]], List[List[float]]]:
    """Parse the block of total forces from the OUTCAR file.

    :param lines: A list of lines containing lines including the TOTAL-FORCE block
    :type lines: list
    :returns: A tuple of position and forces
    :rtype: tuple
    """
    forces = []
    positions = []
    istart = len(lines)
    for idx, line in enumerate(lines):
        if 'TOTAL-FORCE (eV/Angst)' in line:
            istart = idx
        elif idx > istart + 1:
            if not line.startswith(' -----'):  # Still in the block
                values = list(map(float, line.split()))
                positions.append(values[:3])
                forces.append(values[3:])
            else:
                # Reached the end of the block
                break
    return positions, forces


def _parse_neb_outputs(path: Union[str, Path, Any], inputs: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:  # pylint: disable=too-many-branches,too-many-statements
    """Scan for NEB output in the OUTCAR content.

    :param path: Input path or fileobj
    :type path: str or Path or file-like object
    :param inputs: Dictionary where the parsed data should be placed
    :type inputs: dict, optional
    :returns: A dictionary of the parsed data
    :rtype: dict
    """
    inputs = {} if inputs is None else inputs

    if isinstance(path, (str, Path)):
        with open(path, 'r', encoding='utf8') as fobj:
            lines = fobj.readlines()
    # A file-like object
    elif hasattr(path, 'readlines'):
        # Reset seek
        path.seek(0)
        lines = path.readlines()
    else:
        raise ValueError(f"'path' variable is not supported: {path}")

    vtst_data = {'neb_converged': False}
    for idx, line in enumerate(lines):
        if 'NIONS' in line:
            nions = int(line.split()[-1])

        elif 'VTST: version' in line:
            vtst_data['version'] = line.split(':')[1].strip()

        elif 'NEB: Tangent' in line:
            tangents = []
            for isub in range(idx + 2, idx + 99999):
                subline = lines[isub]
                if subline.strip():
                    tangents.append([float(tmp) for tmp in subline.split()])
                else:
                    break
            vtst_data['tangents'] = tangents
        elif 'NEB: forces' in line:
            forces = [float(tmp) for tmp in line.split()[-3:]]
            vtst_data['force_par_spring'] = forces[0]
            vtst_data['force_prep_real'] = forces[1]
            vtst_data['force_dneb'] = forces[2]
        elif 'stress matrix after NEB project' in line:
            stress = []
            for isub in range(idx + 1, idx + 4):
                stress.append([float(tmp) for tmp in lines[isub].split()])
            vtst_data['stress_matrix'] = stress
        elif 'FORCES: max atom' in line:
            forces = [float(tmp) for tmp in line.split()[-2:]]
            vtst_data['force_max_atom'] = forces[0]
            vtst_data['force_rms'] = forces[1]
        elif 'FORCE total and by dimension' in line:
            forces = [float(tmp) for tmp in line.split()[-2:]]
            vtst_data['force_total'] = forces[0]
            vtst_data['force_by_dimension'] = forces[1]
        elif 'Stress total and by dimension' in line:
            forces = [float(tmp) for tmp in line.split()[-2:]]
            vtst_data['stress_total'] = forces[0]
            vtst_data['stress_by_dimension'] = forces[1]
        elif 'OPT: skip step - force has converged' in line:
            vtst_data['neb_converged'] = True
        elif 'energy(sigma->0)' in line:
            tokens = line.split()
            vtst_data['energy_extrapolated'] = float(tokens[-1])
            vtst_data['energy_without_entropy'] = float(tokens[-4])
        elif 'free  energy   TOTEN' in line:
            vtst_data['energy_free'] = float(line.split()[-2])
        elif 'TOTAL-FORCE' in line:
            positions, forces = _parse_force_block(lines[idx : idx + nions + 10])
            inputs['forces'] = np.array(forces)
            inputs['outcar_positions'] = np.array(positions)
        elif 'direct lattice vectors' in line:
            cell = []
            for subline in lines[idx + 1 : idx + 4]:
                cell.append([float(tmp) for tmp in subline.split()[:3]])
            inputs['outcar_cell'] = np.array(cell)

    inputs['neb_data'] = vtst_data
    # Remove the energy keys from the vtst_data dictionary and return them separately
    energies = {}
    for key in ['energy_extrapolated', 'energy_without_entropy', 'energy_free']:
        if key in vtst_data:
            energies[key] = vtst_data.pop(key)
    inputs['total_energies'] = energies

    return inputs
